from __future__ import annotations

from abc import ABC, abstractmethod
from collections.abc import Sequence
from collections.abc import Set as AbstractSet
from dataclasses import dataclass, field
from typing import (
    TYPE_CHECKING,
    Any,
    Generic,
    TypeAlias,
    TypeGuard,
    cast,
)

from typing_extensions import TypeVar, deprecated

from xdsl.ir import (
    Attribute,
    AttributeCovT,
    ParametrizedAttribute,
    TypedAttribute,
)
from xdsl.utils.exceptions import PyRDLError, VerifyException
from xdsl.utils.runtime_final import is_runtime_final

if TYPE_CHECKING:
    from xdsl.irdl import IRDLAttrConstraint


@dataclass
class ConstraintContext:
    """
    Contains the assignment of constraint variables.
    """

    _variables: dict[str, Attribute] = field(default_factory=dict[str, Attribute])
    """The assignment of constraint variables."""

    _range_variables: dict[str, tuple[Attribute, ...]] = field(
        default_factory=dict[str, tuple[Attribute, ...]]
    )
    """The assignment of constraint range variables."""

    _int_variables: dict[str, int] = field(default_factory=dict[str, int])
    """The assignment of constraint int variables."""

    def get_variable(self, key: str) -> Attribute | None:
        return self._variables.get(key)

    def get_range_variable(self, key: str) -> tuple[Attribute, ...] | None:
        return self._range_variables.get(key)

    def get_int_variable(self, key: str) -> int | None:
        return self._int_variables.get(key)

    def set_attr_variable(self, key: str, attr: Attribute):
        self._variables[key] = attr

    def set_range_variable(self, key: str, attrs: tuple[Attribute, ...]):
        self._range_variables[key] = attrs

    def set_int_variable(self, key: str, i: int):
        self._int_variables[key] = i

    @property
    def attr_variables(self) -> AbstractSet[str]:
        return self._variables.keys()

    @property
    def range_variables(self) -> AbstractSet[str]:
        return self._range_variables.keys()

    @property
    def int_variables(self) -> AbstractSet[str]:
        return self._int_variables.keys()

    @deprecated("ConstraintContexts should not be copied")
    def copy(self):
        return ConstraintContext(
            self._variables.copy(),
            self._range_variables.copy(),
            self._int_variables.copy(),
        )

    @deprecated("ConstraintContexts should only be updated by set_* methods")
    def update(self, other: ConstraintContext):
        self._variables.update(other._variables)
        self._range_variables.update(other._range_variables)
        self._int_variables.update(other._int_variables)


_AttributeCovT = TypeVar(
    "_AttributeCovT", bound=Attribute, default=Attribute, covariant=True
)

ConstraintVariableType: TypeAlias = Attribute | Sequence[Attribute] | int
"""
Possible types that a constraint variable can have.
"""


@dataclass(frozen=True)
class AttrConstraint(Generic[AttributeCovT], ABC):
    """Constrain an attribute to a certain value."""

    @abstractmethod
    def verify(
        self,
        attr: Attribute,
        constraint_context: ConstraintContext,
    ) -> None:
        """
        Check if the attribute satisfies the constraint,
        or raise an exception otherwise.
        """
        ...

    def verifies(self, attr: Attribute) -> TypeGuard[AttributeCovT]:
        """
        A helper method to check whether a given attribute matches `self`.
        """
        try:
            self.verify(attr, ConstraintContext())
            return True
        except VerifyException:
            return False

    def variables(self) -> set[str]:
        """
        Returns a set of the variables that can be extracted by this constraint.
        These variables are always expected to be set after running `verify`.
        """
        return set()

    def can_infer(self, var_constraint_names: AbstractSet[str]) -> bool:
        """
        Check if there is enough information to infer the attribute given the
        constraint variables that are already set.
        """
        # By default, we cannot infer anything.
        return False

    def infer(self, context: ConstraintContext) -> AttributeCovT:
        """
        Infer the attribute given the the values for all variables.

        Raises an exception if the attribute cannot be inferred. If `can_infer`
        returns `True` with the given constraint variables, this method should
        not raise an exception.
        """
        raise ValueError(f"Cannot infer attribute from constraint {self}")

    def get_bases(self) -> set[type[Attribute]] | None:
        """
        Get a set of base types that can satisfy this constraint, if there exists
        a finite collection, or None otherwise.
        """
        return None

    def __or__(
        self, value: AttrConstraint[_AttributeCovT], /
    ) -> AttrConstraint[AttributeCovT | _AttributeCovT]:
        if isinstance(value, AnyAttr) or self == value:
            return value  # pyright: ignore[reportReturnType]
        return AnyOf((self, value))

    def __and__(self, value: AttrConstraint, /) -> AttrConstraint[AttributeCovT]:
        if isinstance(value, AnyAttr) or self == value:
            return self
        return AllOf((self, value))  # pyright: ignore[reportReturnType]

    @abstractmethod
    def mapping_type_vars(
        self, type_var_mapping: dict[TypeVar, AttrConstraint]
    ) -> AttrConstraint[AttributeCovT]:
        """
        A helper function to make type vars used in attribute definitions concrete when
        creating constraints for new attributes or operations.
        """
        raise NotImplementedError(
            "Custom constraints must map type vars in nested constraints, if any."
        )


ConstraintVariableTypeT = TypeVar(
    "ConstraintVariableTypeT", bound=ConstraintVariableType
)


TypedAttributeCovT = TypeVar("TypedAttributeCovT", bound=TypedAttribute, covariant=True)
TypedAttributeT = TypeVar("TypedAttributeT", bound=TypedAttribute)


@deprecated("Please use appropriate `AnyOf` constraints instead.")
@dataclass(frozen=True)
class TypedAttributeConstraint(AttrConstraint[TypedAttributeCovT]):
    """
    Constrains the type of a typed attribute.
    """

    attr_constraint: AttrConstraint[TypedAttributeCovT]
    type_constraint: AttrConstraint

    def verify(self, attr: Attribute, constraint_context: ConstraintContext) -> None:
        if not isinstance(attr, TypedAttribute):
            raise VerifyException(f"attribute {attr} expected to be a TypedAttribute")
        self.attr_constraint.verify(attr, constraint_context)
        self.type_constraint.verify(attr.get_type(), constraint_context)

    def variables(self) -> set[str]:
        return self.type_constraint.variables() | self.attr_constraint.variables()

    def can_infer(self, var_constraint_names: AbstractSet[str]) -> bool:
        return self.attr_constraint.can_infer(var_constraint_names)

    def infer(self, context: ConstraintContext) -> TypedAttributeCovT:
        return self.attr_constraint.infer(context)

    def get_bases(self) -> set[type[Attribute]] | None:
        return self.attr_constraint.get_bases()

    def mapping_type_vars(
        self, type_var_mapping: dict[TypeVar, AttrConstraint]
    ) -> TypedAttributeConstraint[TypedAttributeCovT]:  # pyright: ignore[reportDeprecated]
        return TypedAttributeConstraint(  # pyright: ignore[reportDeprecated]
            self.attr_constraint.mapping_type_vars(type_var_mapping),
            self.type_constraint.mapping_type_vars(type_var_mapping),
        )


@dataclass(frozen=True)
class VarConstraint(AttrConstraint[AttributeCovT]):
    """
    Constrain an attribute with the given constraint, and constrain all occurences
    of this constraint (i.e, sharing the same name) to be equal.
    """

    name: str
    """The variable name. All uses of that name refer to the same variable."""

    constraint: AttrConstraint[AttributeCovT]
    """The constraint that the variable must satisfy."""

    def verify(
        self,
        attr: Attribute,
        constraint_context: ConstraintContext,
    ) -> None:
        ctx_attr = constraint_context.get_variable(self.name)
        if ctx_attr is not None:
            if attr != ctx_attr:
                raise VerifyException(
                    f"attribute {constraint_context.get_variable(self.name)} expected from variable "
                    f"'{self.name}', but got {attr}"
                )
        else:
            self.constraint.verify(attr, constraint_context)
            constraint_context.set_attr_variable(self.name, attr)

    def variables(self) -> set[str]:
        return self.constraint.variables() | {self.name}

    def infer(self, context: ConstraintContext) -> AttributeCovT:
        v = context.get_variable(self.name)
        return cast(AttributeCovT, v)

    def can_infer(self, var_constraint_names: AbstractSet[str]) -> bool:
        return self.name in var_constraint_names

    def get_bases(self) -> set[type[Attribute]] | None:
        return self.constraint.get_bases()

    def mapping_type_vars(
        self, type_var_mapping: dict[TypeVar, AttrConstraint]
    ) -> VarConstraint[AttributeCovT]:
        return VarConstraint(
            self.name, self.constraint.mapping_type_vars(type_var_mapping)
        )


@dataclass(frozen=True)
class TypeVarConstraint(AttrConstraint):
    """
    Stores the TypeVar instance used to define a generic ParametrizedAttribute.
    """

    type_var: TypeVar
    """The instance of the TypeVar used in the definition."""

    base_constraint: AttrConstraint
    """Constraint inferred from the base of the TypeVar."""

    def verify(
        self,
        attr: Attribute,
        constraint_context: ConstraintContext,
    ) -> None:
        self.base_constraint.verify(attr, constraint_context)

    def get_bases(self) -> set[type[Attribute]] | None:
        return self.base_constraint.get_bases()

    def mapping_type_vars(
        self, type_var_mapping: dict[TypeVar, AttrConstraint]
    ) -> AttrConstraint:
        res = type_var_mapping.get(self.type_var)
        if res is None:
            raise KeyError(f"Mapping value missing for type var {self.type_var}")
        return res


@dataclass(frozen=True, init=True)
class ConstraintVar:
    """
    Annotation used in PyRDL to define a constraint variable.
    For instance, the following code defines a constraint variable T,
    that can then be used in PyRDL:
    ```python
    T = Annotated[PyRDLConstraint, ConstraintVar("T")]
    ```
    """

    name: str
    """The variable name. All uses of that name refer to the same variable."""


@dataclass(frozen=True)
class EqAttrConstraint(Generic[AttributeCovT], AttrConstraint[AttributeCovT]):
    """Constrain an attribute to be equal to another attribute."""

    attr: AttributeCovT
    """The attribute we want to check equality with."""

    def verify(
        self,
        attr: Attribute,
        constraint_context: ConstraintContext,
    ) -> None:
        if attr != self.attr:
            raise VerifyException(f"Expected attribute {self.attr} but got {attr}")

    def can_infer(self, var_constraint_names: AbstractSet[str]) -> bool:
        return True

    def infer(self, context: ConstraintContext) -> AttributeCovT:
        return self.attr

    def get_bases(self) -> set[type[Attribute]] | None:
        return {type(self.attr)}

    def mapping_type_vars(
        self, type_var_mapping: dict[TypeVar, AttrConstraint]
    ) -> AttrConstraint[AttributeCovT]:
        return self


@dataclass(frozen=True)
class BaseAttr(Generic[AttributeCovT], AttrConstraint[AttributeCovT]):
    """Constrain an attribute to be of a given base type."""

    attr: type[AttributeCovT]
    """The expected attribute base type."""

    def __repr__(self):
        return f"BaseAttr({self.attr.__name__})"

    def verify(
        self,
        attr: Attribute,
        constraint_context: ConstraintContext,
    ) -> None:
        if not isinstance(attr, self.attr):
            raise VerifyException(
                f"{attr} should be of base attribute {self.attr.name}"
            )

    def can_infer(self, var_constraint_names: AbstractSet[str]) -> bool:
        return (
            is_runtime_final(self.attr)
            and issubclass(self.attr, ParametrizedAttribute)
            and not self.attr.get_irdl_definition().parameters
        )

    def infer(self, context: ConstraintContext) -> AttributeCovT:
        assert issubclass(self.attr, ParametrizedAttribute)
        attr = self.attr.new(())
        return attr

    def get_bases(self) -> set[type[Attribute]] | None:
        if is_runtime_final(self.attr):
            return {self.attr}
        return None

    def mapping_type_vars(
        self, type_var_mapping: dict[TypeVar, AttrConstraint]
    ) -> AttrConstraint[AttributeCovT]:
        return self


@deprecated("Please use `irdl_to_attr_constraint` instead")
def attr_constr_coercion(
    attr: AttributeCovT | type[AttributeCovT] | AttrConstraint[AttributeCovT],
) -> AttrConstraint[AttributeCovT]:
    """
    Attributes are coerced into EqAttrConstraints,
    and Attribute types are coerced into BaseAttr.
    """
    from xdsl.irdl import irdl_to_attr_constraint

    return irdl_to_attr_constraint(attr)


@dataclass(frozen=True)
class AnyAttr(AttrConstraint):
    """Constraint that is verified by all attributes."""

    def verify(
        self,
        attr: Attribute,
        constraint_context: ConstraintContext,
    ) -> None:
        pass

    def mapping_type_vars(
        self, type_var_mapping: dict[TypeVar, AttrConstraint]
    ) -> AnyAttr:
        return self

    def __or__(self, value: AttrConstraint[_AttributeCovT], /):
        return self

    def __and__(self, value: AttrConstraint[AttributeCovT], /):
        return value


@dataclass(frozen=True, init=False)
class AnyOf(Generic[AttributeCovT], AttrConstraint[AttributeCovT]):
    """Ensure that an attribute satisfies one of the given constraints."""

    attr_constrs: tuple[AttrConstraint[AttributeCovT], ...]
    """The list of constraints that are checked."""

    _eq_constrs: set[Attribute] = field(hash=False, repr=False)
    _based_constrs: dict[type[Attribute], AttrConstraint[AttributeCovT]] = field(
        hash=False, repr=False
    )

    def __init__(
        self,
        attr_constrs: Sequence[
            AttributeCovT | type[AttributeCovT] | AttrConstraint[AttributeCovT]
        ],
    ):
        from xdsl.irdl import irdl_to_attr_constraint

        constrs: tuple[AttrConstraint[AttributeCovT], ...] = tuple(
            irdl_to_attr_constraint(constr) for constr in attr_constrs
        )

        eq_constrs = set[Attribute]()
        based_constrs = dict[type[Attribute], AttrConstraint[AttributeCovT]]()

        bases = set[Attribute]()
        eq_bases = set[Attribute]()
        for i, c in enumerate(constrs):
            b = c.get_bases()
            if b is None:
                raise PyRDLError(
                    f"Constraint {c} cannot appear in an `AnyOf` constraint as its bases aren't known."
                )

            if not b.isdisjoint(bases):
                raise PyRDLError(
                    f"Constraint {c} shares a base with a non-equality constraint "
                    f"in {set(constrs[0:i])} in `AnyOf` constraint."
                )

            if isinstance(c, EqAttrConstraint):
                eq_constrs.add(c.attr)
                eq_bases |= b
            else:
                if not b.isdisjoint(eq_bases):
                    raise PyRDLError(
                        f"Non-equality constraint {c} shares a base with a constraint "
                        f"in {set(constrs[0:i])} in `AnyOf` constraint."
                    )
                for base in b:
                    based_constrs[base] = c
                bases |= b

        object.__setattr__(
            self,
            "attr_constrs",
            constrs,
        )
        object.__setattr__(
            self,
            "_eq_constrs",
            eq_constrs,
        )
        object.__setattr__(
            self,
            "_based_constrs",
            based_constrs,
        )

    def verify(self, attr: Attribute, constraint_context: ConstraintContext) -> None:
        if attr in self._eq_constrs:
            return
        constr = self._based_constrs.get(attr.__class__)
        if constr is None:
            raise VerifyException(f"Unexpected attribute {attr}")
        constr.verify(attr, constraint_context)

    def __or__(
        self, value: AttrConstraint[_AttributeCovT], /
    ) -> AnyOf[AttributeCovT | _AttributeCovT]:
        return AnyOf((*self.attr_constrs, value))

    def variables(self) -> set[str]:
        if not self.attr_constrs:
            return set()
        variables = self.attr_constrs[0].variables()
        for constr in self.attr_constrs[1:]:
            variables &= constr.variables()
        return variables

    def get_bases(self) -> set[type[Attribute]] | None:
        bases = set[type[Attribute]]()
        for constr in self.attr_constrs:
            b = constr.get_bases()
            if b is None:
                return
            bases |= b
        return bases

    def mapping_type_vars(
        self, type_var_mapping: dict[TypeVar, AttrConstraint]
    ) -> AnyOf[AttributeCovT]:
        return AnyOf(
            tuple(c.mapping_type_vars(type_var_mapping) for c in self.attr_constrs)
        )


@dataclass(frozen=True)
class AllOf(AttrConstraint[AttributeCovT]):
    """Ensure that an attribute satisfies all the given constraints."""

    attr_constrs: tuple[AttrConstraint[AttributeCovT], ...]
    """The list of constraints that are checked."""

    def verify(
        self,
        attr: Attribute,
        constraint_context: ConstraintContext,
    ) -> None:
        exc_bucket: list[VerifyException] = []

        for attr_constr in self.attr_constrs:
            try:
                attr_constr.verify(attr, constraint_context)
            except VerifyException as e:
                exc_bucket.append(e)

        if len(exc_bucket):
            if len(exc_bucket) == 1:
                raise VerifyException(str(exc_bucket[0])) from exc_bucket[0]
            exc_msg = "The following constraints were not satisfied:\n"
            exc_msg += "\n".join([str(e) for e in exc_bucket])
            raise VerifyException(exc_msg)

    def variables(self) -> set[str]:
        vars = set[str]()
        for constr in self.attr_constrs:
            vars |= constr.variables()
        return vars

    def can_infer(self, var_constraint_names: AbstractSet[str]) -> bool:
        return any(
            constr.can_infer(var_constraint_names) for constr in self.attr_constrs
        )

    def infer(self, context: ConstraintContext) -> AttributeCovT:
        for constr in self.attr_constrs:
            if constr.can_infer(context.attr_variables):
                return constr.infer(context)
        raise ValueError("Cannot infer attribute from constraint")

    def get_bases(self) -> set[type[Attribute]] | None:
        bases: set[type[Attribute]] | None = None
        for constr in self.attr_constrs:
            b = constr.get_bases()
            if b is None:
                continue
            if bases is None:
                bases = b
            else:
                bases &= b
        return bases

    def __and__(self, value: AttrConstraint, /) -> AllOf[AttributeCovT]:
        return AllOf((*self.attr_constrs, value))  # pyright: ignore[reportReturnType]

    def mapping_type_vars(
        self, type_var_mapping: dict[TypeVar, AttrConstraint]
    ) -> AllOf[AttributeCovT]:
        return AllOf(
            tuple(c.mapping_type_vars(type_var_mapping) for c in self.attr_constrs)
        )


ParametrizedAttributeT = TypeVar("ParametrizedAttributeT", bound=ParametrizedAttribute)
ParametrizedAttributeCovT = TypeVar(
    "ParametrizedAttributeCovT", bound=ParametrizedAttribute, covariant=True
)


@dataclass(frozen=True, init=False)
class ParamAttrConstraint(
    Generic[ParametrizedAttributeCovT], AttrConstraint[ParametrizedAttributeCovT]
):
    """
    Constrain an attribute to be of a given type,
    and also constrain its parameters with additional constraints.
    """

    base_attr: type[ParametrizedAttributeCovT]
    """The base attribute type."""

    param_constrs: tuple[AttrConstraint, ...]
    """The attribute parameter constraints"""

    def __init__(
        self,
        base_attr: type[ParametrizedAttributeCovT],
        param_constrs: Sequence[IRDLAttrConstraint | None],
    ):
        from xdsl.irdl import irdl_to_attr_constraint

        constrs = tuple(
            irdl_to_attr_constraint(constr) if constr is not None else AnyAttr()
            for constr in param_constrs
        )
        object.__setattr__(self, "base_attr", base_attr)
        object.__setattr__(self, "param_constrs", constrs)

    def __repr__(self):
        return f"ParamAttrConstraint({self.base_attr.__name__}, {repr(self.param_constrs)})"

    def verify(
        self,
        attr: Attribute,
        constraint_context: ConstraintContext,
    ) -> None:
        if not isinstance(attr, self.base_attr):
            raise VerifyException(
                f"{attr} should be of base attribute {self.base_attr.name}"
            )
        parameters = attr.parameters
        if len(self.param_constrs) != len(parameters):
            raise VerifyException(
                f"{len(self.param_constrs)} parameters expected, "
                f"but got {len(parameters)}"
            )
        for idx, param_constr in enumerate(self.param_constrs):
            param_constr.verify(parameters[idx], constraint_context)

    def variables(self) -> set[str]:
        vars = set[str]()
        for constr in self.param_constrs:
            vars |= constr.variables()
        return vars

    def can_infer(self, var_constraint_names: AbstractSet[str]) -> bool:
        return is_runtime_final(self.base_attr) and all(
            constr.can_infer(var_constraint_names) for constr in self.param_constrs
        )

    def infer(self, context: ConstraintContext) -> ParametrizedAttributeCovT:
        params = tuple(constr.infer(context) for constr in self.param_constrs)
        attr = self.base_attr.new(params)
        return attr

    def get_bases(self) -> set[type[Attribute]] | None:
        if is_runtime_final(self.base_attr):
            return {self.base_attr}
        return None

    def mapping_type_vars(
        self, type_var_mapping: dict[TypeVar, AttrConstraint]
    ) -> ParamAttrConstraint[ParametrizedAttributeCovT]:
        return ParamAttrConstraint(
            self.base_attr,
            tuple(c.mapping_type_vars(type_var_mapping) for c in self.param_constrs),
        )

    def __or__(self, value: AttrConstraint[_AttributeCovT], /):
        if (
            not isinstance(value, ParamAttrConstraint)
            or self.base_attr is not cast(ParamAttrConstraint[Any], value).base_attr
        ):
            return super().__or__(value)  # pyright: ignore[reportUnknownArgumentType]
        return ParamAttrConstraint(
            self.base_attr,
            tuple(
                l | r
                for l, r in zip(self.param_constrs, value.param_constrs, strict=True)
            ),
        )


@dataclass(frozen=True, init=False)
class MessageConstraint(AttrConstraint[AttributeCovT]):
    """
    Attach a message to a constraint, to provide more context when the constraint
    is not satisfied.
    """

    constr: AttrConstraint[AttributeCovT]
    message: str

    def __init__(
        self,
        constr: (AttrConstraint[AttributeCovT] | AttributeCovT | type[AttributeCovT]),
        message: str,
    ):
        from xdsl.irdl import irdl_to_attr_constraint

        object.__setattr__(self, "constr", irdl_to_attr_constraint(constr))
        object.__setattr__(self, "message", message)

    def verify(
        self,
        attr: Attribute,
        constraint_context: ConstraintContext,
    ) -> None:
        try:
            return self.constr.verify(attr, constraint_context)
        except VerifyException as e:
            raise VerifyException(
                f"{self.message}\nUnderlying verification failure: {e.args[0]}",
                *e.args[1:],
            )

    def variables(self) -> set[str]:
        return self.constr.variables()

    def get_bases(self) -> set[type[Attribute]] | None:
        return self.constr.get_bases()

    def can_infer(self, var_constraint_names: AbstractSet[str]) -> bool:
        return self.constr.can_infer(var_constraint_names)

    def infer(self, context: ConstraintContext) -> AttributeCovT:
        return self.constr.infer(context)

    def mapping_type_vars(
        self, type_var_mapping: dict[TypeVar, AttrConstraint]
    ) -> MessageConstraint[AttributeCovT]:
        return MessageConstraint(
            self.constr.mapping_type_vars(type_var_mapping), self.message
        )


@dataclass(frozen=True)
class IntConstraint(ABC):
    """Constrain an integer to certain values."""

    @abstractmethod
    def verify(
        self,
        i: int,
        constraint_context: ConstraintContext,
    ) -> None:
        """
        Check if the integer satisfies the constraint, or raise an exception otherwise.
        """
        ...

    def variables(self) -> set[str]:
        """
        Returns a set of the variables that can be extracted by this constraint.
        These variables are always expected to be set after running `verify`.
        """
        return set()

    def can_infer(self, var_constraint_names: AbstractSet[str]) -> bool:
        """
        Check if there is enough information to infer the integer given the
        constraint variables that are already set.
        """
        # By default, we cannot infer anything.
        return False

    def infer(self, context: ConstraintContext) -> int:
        """
        Infer the attribute given the the values for all variables.

        Raises an exception if the attribute cannot be inferred. If `can_infer`
        returns `True` with the given constraint variables, this method should
        not raise an exception.
        """
        raise ValueError(f"Cannot infer integer from constraint {self}")


class AnyInt(IntConstraint):
    """
    Constraint that is verified by all integers.
    """

    def verify(self, i: int, constraint_context: ConstraintContext) -> None:
        pass


@dataclass(frozen=True)
class EqIntConstraint(IntConstraint):
    """Constrain an integer to a value."""

    value: int

    def verify(
        self,
        i: int,
        constraint_context: ConstraintContext,
    ) -> None:
        if self.value != i:
            raise VerifyException(f"Invalid value {i}, expected {self.value}")

    def can_infer(self, var_constraint_names: AbstractSet[str]) -> bool:
        return True

    def infer(self, context: ConstraintContext) -> int:
        return self.value


@dataclass(frozen=True)
class AtLeast(IntConstraint):
    """Constrain an integer to be at least a given value."""

    bound: int
    """The minimum value the integer can take."""

    def verify(self, i: int, constraint_context: ConstraintContext) -> None:
        if i < self.bound:
            raise VerifyException(f"expected integer >= {self.bound}, got {i}")


@dataclass(frozen=True)
class IntVarConstraint(IntConstraint):
    """
    Constrain an integer with the given constraint, and constrain all occurences
    of this constraint (i.e, sharing the same name) to be equal.
    """

    name: str
    """The variable name. All uses of that name refer to the same variable."""

    constraint: IntConstraint
    """The constraint that the variable must satisfy."""

    def verify(
        self,
        i: int,
        constraint_context: ConstraintContext,
    ) -> None:
        if self.name in constraint_context.int_variables:
            if i != constraint_context.get_int_variable(self.name):
                raise VerifyException(
                    f"integer {constraint_context.get_int_variable(self.name)} expected from int variable "
                    f"'{self.name}', but got {i}"
                )
        else:
            self.constraint.verify(i, constraint_context)
            constraint_context.set_int_variable(self.name, i)

    def variables(self) -> set[str]:
        return self.constraint.variables() | {self.name}

    def can_infer(self, var_constraint_names: AbstractSet[str]) -> bool:
        return self.name in var_constraint_names

    def infer(
        self,
        context: ConstraintContext,
    ) -> int:
        v = context.get_int_variable(self.name)
        assert isinstance(v, int)
        return v


@dataclass(frozen=True)
class RangeConstraint(Generic[AttributeCovT], ABC):
    """Constrain a range of attributes to certain values."""

    @abstractmethod
    def verify(
        self,
        attrs: Sequence[Attribute],
        constraint_context: ConstraintContext,
    ) -> None:
        """
        Check if the range satisfies the constraint, or raise an exception otherwise.
        """
        ...

    @abstractmethod
    def verify_length(self, length: int, constraint_context: ConstraintContext) -> None:
        """
        Check if the length of the range satisfies the constraint, or raise an exception otherwise.
        """
        ...

    def variables(self) -> set[str]:
        """
        Returns a set of the variables that can be extracted by this constraint.
        These variables are always expected to be set after running `verify`.
        """
        return set()

    def variables_from_length(self) -> set[str]:
        """
        Returns a set of the variables that can be extracted from the range length by this constraint.
        These variables are always expected to be set after running `verify_length`.
        """
        return set()

    def can_infer(
        self, var_constraint_names: AbstractSet[str], *, length_known: bool
    ) -> bool:
        """
        Check if there is enough information to infer the attribute given the
        constraint variables that are already set, and whether the length of the
        range is known in advance.
        """
        # By default, we cannot infer anything.
        return False

    def infer(
        self, context: ConstraintContext, *, length: int | None
    ) -> Sequence[AttributeCovT]:
        """
        Infer the attribute given the the values for all variables, and possibly
        the length of the range if known.

        Raises an exception if the attribute cannot be inferred. If `can_infer`
        returns `True` with the given constraint variables, this method should
        not raise an exception.
        """
        raise ValueError(f"Cannot infer range from constraint {self}")

    @abstractmethod
    def mapping_type_vars(
        self, type_var_mapping: dict[TypeVar, AttrConstraint]
    ) -> RangeConstraint[AttributeCovT]:
        """
        A helper function to make type vars used in attribute definitions concrete when
        creating constraints for new attributes or operations.
        """
        raise NotImplementedError(
            "Custom constraints must map type vars in nested constraints, if any."
        )

    def of_length(
        self, length_constr: IntConstraint
    ) -> RangeLengthConstraint[AttributeCovT]:
        return RangeLengthConstraint(self, length_constr)


@dataclass(frozen=True)
class RangeLengthConstraint(RangeConstraint[AttributeCovT]):
    """
    Constrain an attribute range with the given length.
    """

    constraint: RangeConstraint[AttributeCovT]
    """The constraint that the variable must satisfy."""

    length: IntConstraint
    """The length that the range must have"""

    def verify(
        self,
        attrs: Sequence[Attribute],
        constraint_context: ConstraintContext,
    ) -> None:
        self.verify_length(len(attrs), constraint_context)
        self.constraint.verify(attrs, constraint_context)

    def verify_length(self, length: int, constraint_context: ConstraintContext) -> None:
        try:
            self.length.verify(length, constraint_context)
        except VerifyException as e:
            raise VerifyException(
                "incorrect length for range variable:\n" + str(e)
            ) from e

    def variables(self) -> set[str]:
        return self.constraint.variables() | self.length.variables()

    def variables_from_length(self) -> set[str]:
        return self.length.variables()

    def can_infer(
        self, var_constraint_names: AbstractSet[str], *, length_known: bool
    ) -> bool:
        length_known = length_known or self.length.can_infer(var_constraint_names)
        return self.constraint.can_infer(
            var_constraint_names, length_known=length_known
        )

    def infer(
        self, context: ConstraintContext, *, length: int | None
    ) -> Sequence[AttributeCovT]:
        if length is None:
            length = self.length.infer(context)
        return self.constraint.infer(context, length=length)

    def mapping_type_vars(
        self, type_var_mapping: dict[TypeVar, AttrConstraint]
    ) -> RangeLengthConstraint[AttributeCovT]:
        return RangeLengthConstraint(
            self.constraint.mapping_type_vars(type_var_mapping), self.length
        )


@dataclass(frozen=True)
class RangeVarConstraint(RangeConstraint[AttributeCovT]):
    """
    Constrain an attribute range with the given constraint, and constrain all occurences
    of this constraint (i.e, sharing the same name) to be equal.
    """

    name: str
    """The variable name. All uses of that name refer to the same variable."""

    constraint: RangeConstraint[AttributeCovT]
    """The constraint that the variable must satisfy."""

    def verify(
        self,
        attrs: Sequence[Attribute],
        constraint_context: ConstraintContext,
    ) -> None:
        ctx_attrs = constraint_context.get_range_variable(self.name)
        if ctx_attrs is not None:
            if attrs != ctx_attrs:
                raise VerifyException(
                    f"attributes {tuple(str(x) for x in ctx_attrs)} expected from range variable "
                    f"'{self.name}', but got {tuple(str(x) for x in attrs)}"
                )
        else:
            self.constraint.verify(attrs, constraint_context)
            constraint_context.set_range_variable(self.name, tuple(attrs))

    def verify_length(self, length: int, constraint_context: ConstraintContext) -> None:
        # It is not possible to fully verify the constraint from just the length, so we don't try.
        pass

    def variables(self) -> set[str]:
        return self.constraint.variables() | {self.name}

    def can_infer(
        self, var_constraint_names: AbstractSet[str], *, length_known: bool
    ) -> bool:
        return self.name in var_constraint_names

    def infer(
        self, context: ConstraintContext, *, length: int | None
    ) -> Sequence[AttributeCovT]:
        v = context.get_range_variable(self.name)
        return cast(Sequence[AttributeCovT], v)

    def mapping_type_vars(
        self, type_var_mapping: dict[TypeVar, AttrConstraint]
    ) -> RangeVarConstraint[AttributeCovT]:
        return RangeVarConstraint(
            self.name, self.constraint.mapping_type_vars(type_var_mapping)
        )


@dataclass(frozen=True)
class RangeOf(RangeConstraint[AttributeCovT]):
    """
    Constrain each element in a range to satisfy a given constraint.
    """

    constr: AttrConstraint[AttributeCovT]

    def verify(
        self,
        attrs: Sequence[Attribute],
        constraint_context: ConstraintContext,
    ) -> None:
        for a in attrs:
            self.constr.verify(a, constraint_context)

    def verify_length(self, length: int, constraint_context: ConstraintContext): ...

    def variables(self) -> set[str]:
        return self.constr.variables()

    def can_infer(
        self, var_constraint_names: AbstractSet[str], *, length_known: bool
    ) -> bool:
        return length_known and self.constr.can_infer(var_constraint_names)

    def infer(
        self,
        context: ConstraintContext,
        *,
        length: int | None,
    ) -> Sequence[AttributeCovT]:
        assert length is not None
        attr = self.constr.infer(context)
        return (attr,) * length

    def mapping_type_vars(
        self, type_var_mapping: dict[TypeVar, AttrConstraint]
    ) -> RangeOf[AttributeCovT]:
        return RangeOf(self.constr.mapping_type_vars(type_var_mapping))


@dataclass(frozen=True)
class SingleOf(RangeConstraint[AttributeCovT]):
    """
    Constrain a range to only contain a single element, which should satisfy a given constraint.
    """

    constr: AttrConstraint[AttributeCovT]

    def verify(
        self,
        attrs: Sequence[Attribute],
        constraint_context: ConstraintContext,
    ) -> None:
        if len(attrs) != 1:
            raise VerifyException(f"Expected a single attribute, got {len(attrs)}")
        self.constr.verify(attrs[0], constraint_context)

    def verify_length(self, length: int, constraint_context: ConstraintContext) -> None:
        if length != 1:
            raise VerifyException(f"Expected a single attribute, got {length}")

    def variables(self) -> set[str]:
        return self.constr.variables()

    def can_infer(
        self, var_constraint_names: AbstractSet[str], *, length_known: int | None
    ) -> bool:
        return self.constr.can_infer(var_constraint_names)

    def infer(
        self, context: ConstraintContext, *, length: int | None
    ) -> Sequence[AttributeCovT]:
        return (self.constr.infer(context),)

    def mapping_type_vars(
        self, type_var_mapping: dict[TypeVar, AttrConstraint]
    ) -> SingleOf[AttributeCovT]:
        return SingleOf(self.constr.mapping_type_vars(type_var_mapping))



#!/usr/bin/env python3
"""Workloads for benchmarking xDSL."""

import random
from math import prod

from xdsl.dialects.arith import AddfOp, AddiOp, ConstantOp
from xdsl.dialects.builtin import (
    DenseIntOrFPElementsAttr,
    Float32Type,
    FunctionType,
    IntegerAttr,
    ModuleOp,
    TensorType,
    i8,
    i32,
)
from xdsl.dialects.func import FuncOp, ReturnOp
from xdsl.dialects.test import TestOp
from xdsl.ir import Operation

RANDOM_SEED = 0
HEX_CHARS = "0123456789ABCDEF"


class WorkloadBuilder:
    """A helper class to programmatically build synthetic workloads."""

    @classmethod
    def wrap_module(cls, ops: list[str]) -> str:
        """Wrap a list of operations as a module."""
        joined_ops = "\n  ".join(ops)
        workload = f'"builtin.module"() ({{\n  {joined_ops}\n}}) : () -> ()'
        return workload

    @classmethod
    def empty(cls) -> str:
        """Generate an empty module."""
        return WorkloadBuilder.wrap_module([])

    @classmethod
    def constant_folding_module(cls, size: int = 100) -> ModuleOp:
        """Generate a constant folding workload of a given size.

        The output of running the command
        `print(WorkloadBuilder().constant_folding_module(size=5))` is shown
        below:

        ```mlir
        "builtin.module"() ({
            %0 = "arith.constant"() {"value" = 865 : i32} : () -> i32
            %1 = "arith.constant"() {"value" = 395 : i32} : () -> i32
            %2 = "arith.addi"(%1, %0) : (i32, i32) -> i32
            %3 = "arith.constant"() {"value" = 777 : i32} : () -> i32
            %4 = "arith.addi"(%3, %2) : (i32, i32) -> i32
            %5 = "arith.constant"() {"value" = 912 : i32} : () -> i32
            "test.op"(%4) : (i32) -> ()
        }) : () -> ()
        ```
        """
        assert size >= 0
        random.seed(RANDOM_SEED)
        ops: list[Operation] = []
        ops.append(ConstantOp(IntegerAttr(random.randint(1, 1000), i32)))
        for i in range(1, size + 1):
            if i % 2 == 0:
                ops.append(AddiOp(ops[i - 1], ops[i - 2]))
            else:
                ops.append(ConstantOp(IntegerAttr(random.randint(1, 1000), i32)))
        ops.append(TestOp([ops[(size // 2) * 2]]))
        return ModuleOp(ops)

    @classmethod
    def constant_folding(cls, size: int = 100) -> str:
        """Generate a constant folding workload of a given size."""
        return str(cls.constant_folding_module(size=size))

    @classmethod
    def large_dense_attr(cls, x: int = 1024, y: int = 1024) -> str:
        """Get the MLIR text representation of a large dense attr."""
        return str(cls.large_dense_attr_module(x=x, y=y))

    @classmethod
    def large_dense_attr_module(cls, x: int = 1024, y: int = 1024) -> ModuleOp:
        """Get the MLIR text representation of a large dense attr.

        An example of running
        `print(WorkloadBuilder().large_dense_attr_module(x=3, y=3))`
        is as follows:

        ```mlir
        "builtin.module"() ({
            %0 = "arith.constant"() <{value = dense<[
                [69, 87, -108], [4, 120, 79], [27, 116, 55]
            ]> : tensor<3x3xi8>}> : () -> tensor<3x3xi8>
        }) : () -> ()
        """
        assert x >= 0
        assert y >= 0
        random.seed(RANDOM_SEED)
        dense_attr = [random.randint(-128, 128) for _ in range(x * y)]
        tensor_type = TensorType(element_type=i8, shape=[x, y])
        return ModuleOp(
            [
                ConstantOp(
                    DenseIntOrFPElementsAttr.from_list(
                        type=tensor_type, data=dense_attr
                    )
                )
            ]
        )

    @classmethod
    def large_dense_attr_hex(cls, x: int = 1024, y: int = 1024) -> str:
        """Get the MLIR hex representation of a large dense attr.

        An example of running `WorkloadBuilder().large_dense_attr(x=3, y=3)`
        is as follows:

        ```mlir
        "builtin.module"() ({
            %0 = "arith.constant"() <{
                value = dense<0xCD18FC9FB649438493> : tensor<3x3xi8>
            }> : () -> tensor<3x3xi8>
        }) : () -> ()
        """
        assert x >= 0
        assert y >= 0
        random.seed(RANDOM_SEED)
        # Each dense attr item is a byte = 2 hex chars
        dense_attr_hex = "".join(random.choice(HEX_CHARS) for _ in range(x * y * 2))
        ops = [
            (
                '%0 = "arith.constant"() '
                f'<{{value = dense<"0x{dense_attr_hex}"> '
                f": tensor<{x}x{y}xi8>}}> : () -> tensor<{x}x{y}xi8>"
            )
        ]
        return WorkloadBuilder.wrap_module(ops)

    @classmethod
    def large_constant_tensor(
        cls,
        tensor_shape: tuple[int, ...],
        num_add_ops: int = 10,
    ) -> ModuleOp:
        """Create a module with a function that adds multiple large constant tensors.

        Args:
            tensor_shape: The shape of the constant tensors to be created.
            num_add_ops: The number of addition operations to perform with the constant tensors.

        Returns:
            The created module containing the function with addition operations.
        """
        assert num_add_ops >= 0
        random.seed(RANDOM_SEED)
        tensor_type = TensorType(shape=tensor_shape, element_type=Float32Type())
        function_type = FunctionType.from_lists(
            inputs=[tensor_type], outputs=[tensor_type]
        )

        func_op = FuncOp(name="main", function_type=function_type)

        input_val = func_op.args[0]
        output_val = input_val

        for _ in range(num_add_ops):
            # Create a constant tensor with random values
            dense_elements_attr = DenseIntOrFPElementsAttr.from_list(
                type=tensor_type,
                data=[random.random() for _ in range(prod(tensor_shape))],
            )
            const_op = ConstantOp(value=dense_elements_attr, value_type=tensor_type)
            func_op.body.block.add_op(const_op)

            # Add the constant to the input value
            add_op = AddfOp(
                operand1=input_val, operand2=const_op, result_type=tensor_type
            )
            func_op.body.block.add_op(add_op)

            # Update input_val to the result of the addition
            input_val = add_op.result
            output_val = add_op.result

        return_op = ReturnOp(output_val)
        func_op.body.block.add_op(return_op)

        module_op = ModuleOp([func_op])
        module_op.verify()
        return module_op



import os
from dataclasses import dataclass

from xdsl.context import Context
from xdsl.dialects import builtin, pdl_interp
from xdsl.interpreter import Interpreter
from xdsl.interpreters.eqsat_pdl_interp import EqsatPDLInterpFunctions
from xdsl.parser import Parser
from xdsl.passes import ModulePass
from xdsl.pattern_rewriter import PatternRewriterListener, PatternRewriteWalker
from xdsl.traits import SymbolTable
from xdsl.transforms.apply_pdl_interp import PDLInterpRewritePattern

_DEFAULT_MAX_ITERATIONS = 20
"""Default number of times to iterate over the module."""


def apply_eqsat_pdl_interp(
    op: builtin.ModuleOp,
    ctx: Context,
    pdl_interp_module: builtin.ModuleOp,
    max_iterations: int = _DEFAULT_MAX_ITERATIONS,
):
    matcher = SymbolTable.lookup_symbol(pdl_interp_module, "matcher")
    assert isinstance(matcher, pdl_interp.FuncOp)
    assert matcher is not None, "matcher function not found"

    # Initialize interpreter and implementations once
    interpreter = Interpreter(pdl_interp_module)
    implementations = EqsatPDLInterpFunctions(ctx)
    implementations.populate_known_ops(op)
    interpreter.register_implementations(implementations)
    rewrite_pattern = PDLInterpRewritePattern(matcher, interpreter, implementations)

    listener = PatternRewriterListener()
    listener.operation_modification_handler.append(implementations.modification_handler)
    walker = PatternRewriteWalker(rewrite_pattern, apply_recursively=False)
    walker.listener = listener

    for _i in range(max_iterations):
        # Register matches by walking the module
        walker.rewrite_module(op)
        # Execute all pending rewrites that were aggregated during matching
        implementations.execute_pending_rewrites(interpreter)

        if not implementations.worklist:
            break

        implementations.rebuild()


@dataclass(frozen=True)
class ApplyEqsatPDLInterpPass(ModulePass):
    name = "apply-eqsat-pdl-interp"

    pdl_interp_file: str | None = None
    max_iterations: int = _DEFAULT_MAX_ITERATIONS
    """Maximum number of iterations to run, default 20."""

    def apply(self, ctx: Context, op: builtin.ModuleOp) -> None:
        if self.pdl_interp_file is not None:
            assert os.path.exists(self.pdl_interp_file)
            with open(self.pdl_interp_file) as f:
                pdl_interp_module_str = f.read()
                parser = Parser(ctx, pdl_interp_module_str)
                pdl_interp_module = parser.parse_module()
        else:
            pdl_interp_module = op

        apply_eqsat_pdl_interp(op, ctx, pdl_interp_module, self.max_iterations)



import pytest

from xdsl.builder import Builder, ImplicitBuilder
from xdsl.dialects import arith, func, scf
from xdsl.dialects.builtin import IndexType, ModuleOp, i1, i32
from xdsl.interpreter import Interpreter, OpCounter
from xdsl.interpreters.arith import ArithFunctions
from xdsl.interpreters.func import FuncFunctions
from xdsl.interpreters.scf import ScfFunctions
from xdsl.ir import BlockArgument

index = IndexType()


def sum_to_for_fn(n: int) -> int:
    """
    Python implementation of sum_to_for_op
    """
    result = 0
    for i in range(0, n, 1):
        result += i
    return result


@ModuleOp
@Builder.implicit_region
def sum_to_for_op():
    with ImplicitBuilder(func.FuncOp("sum_to", ((index,), (index,))).body) as (ub,):
        lb = arith.ConstantOp.from_int_and_width(0, index)
        step = arith.ConstantOp.from_int_and_width(1, index)
        initial = arith.ConstantOp.from_int_and_width(0, index)

        @Builder.implicit_region((index, index))
        def for_loop_region(args: tuple[BlockArgument, ...]):
            (i, acc) = args
            res = arith.AddiOp(i, acc)
            scf.YieldOp(res)

        result = scf.ForOp(lb, ub, step, (initial,), for_loop_region)
        func.ReturnOp(result)


def scf_interp(module_op: ModuleOp, func_name: str, n: int) -> int:
    module_op.verify()
    interpreter = Interpreter(module_op)
    interpreter.register_implementations(ScfFunctions())
    interpreter.register_implementations(FuncFunctions())
    interpreter.register_implementations(ArithFunctions())
    (result,) = interpreter.call_op(func_name, (n,))
    return result


@pytest.mark.parametrize("n,res", [(0, 0), (1, 0), (2, 1), (3, 3), (4, 6), (5, 10)])
def test_sum_to(n: int, res: int):
    assert res == scf_interp(sum_to_for_op, "sum_to", n)


def test_if():
    @ModuleOp
    @Builder.implicit_region
    def module_op():
        with ImplicitBuilder(func.FuncOp("indicator", ((i1,), (i32,))).body) as (cond,):

            @Builder.implicit_region
            def true_region():
                one = arith.ConstantOp.from_int_and_width(1, 32)
                scf.YieldOp(one)

            @Builder.implicit_region
            def false_region():
                zero = arith.ConstantOp.from_int_and_width(0, 32)
                scf.YieldOp(zero)

            result = scf.IfOp(cond, (i32,), true_region, false_region)

            func.ReturnOp(result)

    assert scf_interp(module_op, "indicator", True) == 1
    assert scf_interp(module_op, "indicator", False) == 0


def test_tracer():
    tracer = OpCounter()
    interpreter = Interpreter(sum_to_for_op.clone(), listeners=(tracer,))
    interpreter.register_implementations(ScfFunctions())
    interpreter.register_implementations(FuncFunctions())
    interpreter.register_implementations(ArithFunctions())
    (result,) = interpreter.call_op("sum_to", (5,))

    assert result == 10
    assert dict(tracer.ops) == {
        "arith.constant": 3,
        "scf.for": 1,
        "scf.yield": 5,
        "arith.addi": 5,
        "func.return": 1,
    }



import abc
from dataclasses import dataclass, field
from typing import Generic

from typing_extensions import TypeVar

T = TypeVar("T")
TCov = TypeVar("TCov", covariant=True)
TCon = TypeVar("TCon", contravariant=True)


class ReadableStream(Generic[TCov], abc.ABC):
    """
    Abstract base class for readable stream interpreter model objects.
    """

    @abc.abstractmethod
    def read(self) -> TCov:
        raise NotImplementedError()


class WritableStream(Generic[TCon], abc.ABC):
    """
    Abstract base class for readable stream interpreter model objects.
    """

    @abc.abstractmethod
    def write(self, value: TCon) -> None:
        raise NotImplementedError()


@dataclass
class Nats(ReadableStream[int]):
    """
    A stream designed for testing, outputs the next natural number each time it's read.
    """

    index = 0

    def read(self) -> int:
        self.index += 1
        return self.index


@dataclass
class Acc(WritableStream[int]):
    """
    A stream designed for testing, appends the next natural number written.
    """

    values: list[int] = field(default_factory=list[int])

    def write(self, value: int) -> None:
        return self.values.append(value)



from dataclasses import dataclass, field

from xdsl.context import Context
from xdsl.dialects import memref, memref_stream
from xdsl.dialects.builtin import ArrayAttr, ModuleOp
from xdsl.ir import Block, Region
from xdsl.passes import ModulePass
from xdsl.pattern_rewriter import (
    PatternRewriter,
    PatternRewriteWalker,
    RewritePattern,
    op_type_rewrite_pattern,
)
from xdsl.rewriter import InsertPoint


@dataclass
class StreamifyGenericOpPattern(RewritePattern):
    streams: int = field()

    @op_type_rewrite_pattern
    def match_and_rewrite(
        self, op: memref_stream.GenericOp, rewriter: PatternRewriter
    ) -> None:
        if any(
            isinstance(
                operand.type,
                memref_stream.ReadableStreamType | memref_stream.WritableStreamType,
            )
            for operand in op.operands
        ):
            # Already streamified
            return

        init_indices = set(index.data for index in op.init_indices)

        # Can only stream memrefs that are not inout
        input_count = len(op.inputs)
        streamable_input_indices = tuple(
            (index, arg.type)
            for index, (i, arg) in enumerate(
                zip(op.inputs, op.body.block.args[:input_count])
            )
            if isinstance(i_type := i.type, memref.MemRefType) and arg.uses
            if i_type.get_shape()
        )
        streamable_output_indices = tuple(
            (index, arg.type)
            for index, (o, arg) in enumerate(
                zip(op.outputs, op.body.block.args[input_count:])
            )
            if isinstance(o_type := o.type, memref.MemRefType)
            if index in init_indices or not arg.uses
            if o_type.get_shape()
        )
        if not streamable_input_indices and not streamable_output_indices:
            # No memrefs to convert to streams
            return
        # We might want to pick which memref to stream by iteration count in the future
        streamed_input_indices = streamable_input_indices[: self.streams]
        streamed_output_indices = streamable_output_indices[
            : self.streams - len(streamed_input_indices)
        ]
        streamed_operand_indices = streamed_input_indices + tuple(
            (index + input_count, el_type) for index, el_type in streamed_output_indices
        )
        input_el_types = tuple(el_type for _, el_type in streamed_input_indices)
        output_el_types = tuple(el_type for _, el_type in streamed_output_indices)
        input_stream_types = tuple(
            memref_stream.ReadableStreamType(el_type) for el_type in input_el_types
        )
        output_stream_types = tuple(
            memref_stream.WritableStreamType(el_type) for el_type in output_el_types
        )

        # input patterns are never unnested
        input_patterns = tuple(
            memref_stream.StridePattern(
                op.bounds,
                indexing_map,
            )
            for index, _ in streamable_input_indices
            if (indexing_map := op.indexing_maps.data[index])
        )
        # output patterns never contain iteration dimensions
        output_patterns = tuple(
            memref_stream.StridePattern(
                ArrayAttr(
                    tuple(
                        bound
                        for iterator_type, bound in zip(
                            op.iterator_types, op.bounds.data
                        )
                        if iterator_type.data != memref_stream.IteratorType.REDUCTION
                    )
                ),
                indexing_map,
            )
            for output_index, _ in streamed_output_indices
            if (indexing_map := op.indexing_maps.data[output_index + input_count])
        )

        patterns = ArrayAttr(input_patterns + output_patterns)
        rewriter.insert_op_before_matched_op(
            streaming_region_op := memref_stream.StreamingRegionOp(
                tuple(op.inputs[index] for index, _ in streamed_input_indices),
                tuple(op.outputs[index] for index, _ in streamable_output_indices),
                patterns,
                Region(Block(arg_types=input_stream_types + output_stream_types)),
            )
        )
        new_body = streaming_region_op.body.block
        new_operands = list(op.operands[: len(op.inputs) + len(op.outputs)])
        for stream_index, (index, _) in enumerate(streamed_operand_indices):
            new_operands[index] = new_body.args[stream_index]

        rewriter.insert_op(
            memref_stream.GenericOp(
                new_operands[:input_count],
                new_operands[input_count:],
                op.inits,
                rewriter.move_region_contents_to_new_regions(op.body),
                op.indexing_maps,
                op.iterator_types,
                op.bounds,
                op.init_indices,
                op.doc,
                op.library_call,
            ),
            InsertPoint.at_end(new_body),
        )
        rewriter.erase_matched_op()


@dataclass(frozen=True)
class MemRefStreamifyPass(ModulePass):
    """
    Converts a memref generic on memrefs to a memref generic on streams, by moving it
    into a streaming region.
    """

    name = "memref-streamify"

    streams: int = field(default=3)

    def apply(self, ctx: Context, op: ModuleOp) -> None:
        PatternRewriteWalker(
            StreamifyGenericOpPattern(self.streams),
            apply_recursively=False,
        ).rewrite_module(op)



from dataclasses import dataclass

from jax._src.core import ClosedJaxpr

from xdsl.builder import Builder, InsertPoint
from xdsl.dialects.builtin import FunctionType, ModuleOp, TensorType, f32
from xdsl.dialects.func import FuncOp, ReturnOp
from xdsl.ir import Block, Region


class IRGenError(Exception):
    pass


@dataclass(init=False)
class IRGen:
    """
    Implementation of a simple MLIR emission from the jaxpr.

    This will emit operations that are specific to the Jax language, preserving
    the semantics of the language and (hopefully) allow to perform accurate
    analysis and transformation based on these high level semantics.
    """

    module: ModuleOp
    """A "module" matches a jaxpr source file: containing a list of functions."""

    builder: Builder

    """
    The builder is a helper class to create IR inside a function. The builder
    is stateful, in particular it keeps an "insertion point": this is where
    the next operations will be introduced."""

    def __init__(self):
        # We create an empty MLIR module and codegen functions one at a time and
        # add them to the module.
        self.module = ModuleOp([])
        self.builder = Builder(InsertPoint.at_end(self.module.body.blocks[0]))

    def ir_gen_module(self, jaxpr: ClosedJaxpr) -> ModuleOp:
        """
        Public API: convert the jaxpr (source file) to an MLIR
        Module operation."""
        parent_builder = self.builder

        assert len(jaxpr.jaxpr.invars) == 1
        assert len(jaxpr.jaxpr.outvars) == 1
        for arg in jaxpr.jaxpr.invars:
            assert arg.aval.dtype == "float32"

        inputVars = jaxpr.jaxpr.invars[0]
        outputVars = jaxpr.jaxpr.outvars[0]

        input_types = [
            TensorType(f32, [inputVars.aval.size])
            for _ in range(len(jaxpr.jaxpr.invars))
        ]
        output_types = [
            TensorType(f32, [outputVars.aval.size])
            for _ in range(len(jaxpr.jaxpr.outvars))
        ]

        block = Block(arg_types=input_types)
        self.builder = Builder(InsertPoint.at_end(block))

        func_type = FunctionType.from_lists(input_types, output_types)

        for _ in jaxpr.jaxpr.eqns:
            raise NotImplementedError("jax equation not implemented")

        assert inputVars == outputVars

        return_ssa = block.args[0]
        self.builder.insert(ReturnOp(return_ssa))

        self.builder = parent_builder

        self.builder.insert(
            FuncOp("main", func_type, Region(block), visibility="public")
        )

        return self.module



import abc
from collections.abc import Iterator, Sequence
from typing import NamedTuple

from xdsl.backend.register_allocator import BlockAllocator
from xdsl.backend.register_type import RegisterType
from xdsl.ir import Operation, Region, SSAValue
from xdsl.irdl import traits_def
from xdsl.traits import OpTrait
from xdsl.utils.exceptions import VerifyException


class RegisterAllocatableOperation(Operation, abc.ABC):
    """
    An abstract base class for operations that can be processed during register
    allocation.
    """

    def iter_used_registers(self) -> Iterator[RegisterType]:
        """
        The registers whose contents may be overwritten when executing this operation.
        By default returns the types of operands and results that are allocated
        registers.
        """
        return (
            val.type
            for vals in (self.operands, self.results)
            for val in vals
            if isinstance(val.type, RegisterType) and val.type.is_allocated
        )

    @abc.abstractmethod
    def allocate_registers(self, allocator: BlockAllocator) -> None:
        """
        Allocate registers for this operation.
        """

    @staticmethod
    def iter_all_used_registers(
        region: Region,
    ) -> Iterator[RegisterType]:
        """
        All used registers of all operations within a region.
        """
        return (
            reg
            for op in region.walk()
            if isinstance(op, RegisterAllocatableOperation)
            for reg in op.iter_used_registers()
        )


class RegisterConstraints(NamedTuple):
    """
    Values used by an instruction.
    A collection of operations in `inouts` represents the constraint that they must be
    allocated to the same register.
    """

    ins: Sequence[SSAValue]
    outs: Sequence[SSAValue]
    inouts: Sequence[Sequence[SSAValue]]


class HasRegisterConstraintsTrait(OpTrait):
    """
    Trait that verifies that the operation implements HasRegisterConstraints, and that
    its inout operands are used only once.
    Using an inout operand more than once breaks SSA, as the register will hold an
    unexpected value after being mutated by this operation.
    """

    def verify(self, op: Operation) -> None:
        if not isinstance(op, HasRegisterConstraints):
            raise VerifyException(
                f"Operation {op.name} is not a subclass of {HasRegisterConstraints.__name__}."
            )

        for o, _ in op.get_register_constraints().inouts:
            if not o.has_one_use():
                raise VerifyException(
                    f"Inout register operand at index {op.operands.index(o)} used more than once."
                )


class HasRegisterConstraints(RegisterAllocatableOperation, abc.ABC):
    """
    Abstract superclass for operations corresponding to assembly, with registers used
    as in, out, or inout registers.
    Inout registers must only be used once.
    """

    traits = traits_def(HasRegisterConstraintsTrait())

    @abc.abstractmethod
    def get_register_constraints(self) -> RegisterConstraints:
        """
        The values with register types used by this operation, for use in register
        allocation.
        """
        raise NotImplementedError()

    def allocate_registers(self, allocator: BlockAllocator) -> None:
        ins, outs, inouts = self.get_register_constraints()

        # Allocate registers to inout operand groups since they are defined further up
        # in the use-def SSA chain
        for operand_group in inouts:
            allocator.allocate_values_same_reg(operand_group)

        for result in outs:
            # Allocate registers to result if not already allocated
            if (new_result := allocator.allocate_value(result)) is not None:
                result = new_result
            allocator.free_value(result)

        # Allocate registers to operands since they are defined further up
        # in the use-def SSA chain
        for operand in ins:
            allocator.allocate_value(operand)



from xdsl.context import Context
from xdsl.dialects import builtin, eqsat, func
from xdsl.ir import Block
from xdsl.passes import ModulePass
from xdsl.pattern_rewriter import (
    GreedyRewritePatternApplier,
    PatternRewriter,
    PatternRewriteWalker,
    RewritePattern,
    op_type_rewrite_pattern,
)
from xdsl.rewriter import InsertPoint, Rewriter
from xdsl.utils.exceptions import DiagnosticException


def insert_eclass_ops(block: Block):
    # Insert eqsat.eclass for each operation
    for op in block.ops:
        results = op.results

        # Skip special ops such as return ops
        if isinstance(op, func.ReturnOp):
            continue

        if len(results) != 1:
            raise DiagnosticException("Ops with non-single results not handled")

        eclass_op = eqsat.EClassOp(results[0])
        insertion_point = InsertPoint.after(op)
        Rewriter.insert_op(eclass_op, insertion_point)
        results[0].replace_by_if(
            eclass_op.results[0], lambda u: not isinstance(u.operation, eqsat.EClassOp)
        )

    # Insert eqsat.eclass for each arg
    for arg in block.args:
        eclass_op = eqsat.EClassOp(arg)
        insertion_point = InsertPoint.at_start(block)
        Rewriter.insert_op(eclass_op, insertion_point)
        arg.replace_by_if(
            eclass_op.results[0], lambda u: not isinstance(u.operation, eqsat.EClassOp)
        )


class InsertEclassOps(RewritePattern):
    """
    Inserts a `eqsat.eclass` after each operation except module op and function op.
    """

    @op_type_rewrite_pattern
    def match_and_rewrite(self, op: func.FuncOp, rewriter: PatternRewriter):
        insert_eclass_ops(op.body.block)


class EqsatCreateEclassesPass(ModulePass):
    """
    Create initial eclasses from an MLIR program.

    Input example:
       ```mlir
       func.func @test(%a : index, %b : index) -> (index) {
           %c = arith.addi %a, %b : index
           func.return %c : index
       }
       ```
    Output example:
        ```mlir
        func.func @test(%a : index, %b : index) -> (index) {
            %a_eq = eqsat.eclass %a : index
            %b_eq = eqsat.eclass %b : index
            %c = arith.addi %a_eq, %b_eq : index
            %c_eq = eqsat.eclass %c : index
            func.return %c_eq : index
        }
        ```
    """

    name = "eqsat-create-eclasses"

    def apply(self, ctx: Context, op: builtin.ModuleOp) -> None:
        PatternRewriteWalker(
            GreedyRewritePatternApplier([InsertEclassOps()]),
            apply_recursively=False,
        ).rewrite_module(op)



from dataclasses import dataclass, field

from xdsl.context import Context
from xdsl.dialects.builtin import ModuleOp
from xdsl.ir import Operation, Region, SSAValue
from xdsl.ir.post_order import PostOrderIterator
from xdsl.passes import ModulePass
from xdsl.pattern_rewriter import (
    PatternRewriter,
    PatternRewriterListener,
    PatternRewriteWalker,
    RewritePattern,
)
from xdsl.traits import (
    IsTerminator,
    MemoryEffectKind,
    SymbolOpInterface,
    get_effects,
)


def would_be_trivially_dead(op: Operation):
    """
    Returns if the operation would be dead if all its results were dead.
    """
    return (
        not op.get_trait(IsTerminator)
        and (not op.get_trait(SymbolOpInterface))
        and result_only_effects(op)
    )


def is_trivially_dead(op: Operation):
    """
    Returns if the operation has no observable effect.
    """
    return all(
        result.first_use is None for result in op.results
    ) and would_be_trivially_dead(op)


def result_only_effects(rootOp: Operation) -> bool:
    """
    Returns if we can ensure the operation would have no observable effect beyond its
    returned values.

    cf MLIR's WouldOpBeTriviallyDead:
    https://mlir.llvm.org/doxygen/namespacemlir.html#a655db45ed8c23d04d5ed5ee0abe041ad

    We have one key difference here:
    - MLIR discard any allocation from an operation on its own result for this analysis
    - xDSL discard any allocation effect of any nested operation on any value defined
    by the root operation or its children.
    """
    effects = get_effects(rootOp)
    # If the operation has unknown effect, we safely assume it has observable ones
    return effects is not None and all(
        # Read-only effect will not affect other operations
        e.kind == MemoryEffectKind.READ
        # Allocation of values defined by this operation or its children will not
        # affect other operations
        or (
            e.kind == MemoryEffectKind.ALLOC
            and isinstance(v := e.value, SSAValue)
            and rootOp.is_ancestor(v.owner)
        )
        for e in effects
    )


class RemoveUnusedOperations(RewritePattern):
    """
    Removes operations annotated with the `Pure` trait, where results have no uses.
    """

    def match_and_rewrite(self, op: Operation, rewriter: PatternRewriter):
        if is_trivially_dead(op) and op.parent is not None:
            rewriter.erase_op(op)


def dce(op: ModuleOp):
    """
    Removes operations annotated with the `Pure` trait, where results have no uses.
    Modifies input module in-place.
    """
    walker = PatternRewriteWalker(
        RemoveUnusedOperations(), apply_recursively=True, walk_reverse=True
    )
    walker.rewrite_module(op)


@dataclass
class LiveSet:
    changed: bool = field(default=True)  # Force first iteration to happen
    _live_ops: set[Operation] = field(default_factory=set[Operation])

    def is_live(self, op: Operation) -> bool:
        return op in self._live_ops

    def set_live(self, op: Operation):
        if not self.is_live(op):
            self.changed = True
            self._live_ops.add(op)

    def propagate_op_liveness(self, op: Operation):
        for region in op.regions:
            self.propagate_region_liveness(region)

        if self.is_live(op):
            return

        if not would_be_trivially_dead(op):
            self.set_live(op)
            return

        if any(
            self.is_live(use.operation) for result in op.results for use in result.uses
        ):
            self.set_live(op)

    def propagate_region_liveness(self, region: Region):
        first = region.first_block
        if first is None:
            return
        for block in PostOrderIterator(first):
            # Process operations in reverse order to speed convergence
            for operation in reversed(block.ops):
                self.propagate_op_liveness(operation)

    def delete_dead(self, region: Region, listener: PatternRewriterListener | None):
        first = region.first_block
        if first is None:
            return

        for block in reversed(region.blocks):
            if not any(self.is_live(op) for op in block.ops) and block != first:
                # If block is not the entry block and has no live ops then delete it
                self.changed = True
                region.erase_block(block, safe_erase=False)
                continue

            for operation in reversed(block.ops):
                if not self.is_live(operation):
                    self.changed = True
                    if listener is not None:
                        listener.handle_operation_removal(operation)
                    block.erase_op(operation, safe_erase=False)
                else:
                    for r in operation.regions:
                        self.delete_dead(r, listener)


def region_dce(region: Region, listener: PatternRewriterListener | None = None) -> bool:
    live_set = LiveSet()

    while live_set.changed:
        live_set.changed = False
        live_set.propagate_region_liveness(region)

    live_set.delete_dead(region, listener)
    return live_set.changed


class DeadCodeElimination(ModulePass):
    name = "dce"

    def apply(self, ctx: Context, op: ModuleOp) -> None:
        region_dce(op.body)



from xdsl.context import Context
from xdsl.dialects.builtin import Builtin, ModuleOp
from xdsl.dialects.test import Test, TestOp
from xdsl.parser import Parser

test_prog = """
"builtin.module"() ({
  %0 = "test.op"() : () -> i32
  %1 = "test.op"(%0, %0) : (i32, i32) -> i32
}) : () -> ()
"""


def test_main():
    ctx = Context()
    ctx.load_dialect(Builtin)
    ctx.load_dialect(Test)

    parser = Parser(ctx, test_prog)
    module = parser.parse_op()

    module.verify()
    assert isinstance(module, ModuleOp)

    op1, op2 = list(module.ops)
    assert set((use.operation, use.index) for use in op1.results[0].uses) == {
        (op2, 0),
        (op2, 1),
    }
    assert set(op2.results[0].uses) == set()


def test_uses_methods():
    ctx = Context()
    ctx.load_dialect(Builtin)
    ctx.load_dialect(Test)

    test_prog = """
    "builtin.module"() ({
      %0 = "test.op"() : () -> i32
      %1 = "test.op"(%0, %0) : (i32, i32) -> i32
      %2 = "test.op"(%1) : (i32) -> i32
    }) : () -> ()
    """

    parser = Parser(ctx, test_prog)
    module = parser.parse_op()

    module.verify()
    assert isinstance(module, ModuleOp)

    op0, op1, op2 = list(module.ops)
    res0, res1, res2 = op0.results[0], op1.results[0], op2.results[0]
    assert not res0.has_one_use()
    assert res1.has_one_use()
    assert not res2.has_one_use()

    assert res0.has_more_than_one_use()
    assert not res1.has_more_than_one_use()
    assert not res2.has_more_than_one_use()

    assert res0.get_unique_use() is None
    assert (use := res1.get_unique_use()) is not None
    assert use.operation == op2
    assert use.index == 0
    assert res2.get_unique_use() is None

    assert res0.get_user_of_unique_use() is None
    assert res1.get_user_of_unique_use() == op2
    assert res2.get_user_of_unique_use() is None


test_prog_blocks = """
"test.op"() ({
  "test.termop"() [^0, ^1] : () -> ()
^0:
  "test.termop"()[^2] : () -> ()
^1:
  "test.termop"()[^2] : () -> ()
^2:
  "test.termop"() : () -> ()
}) : () -> ()
"""


def test_predecessor():
    ctx = Context()
    ctx.load_dialect(Builtin)
    ctx.load_dialect(Test)

    parser = Parser(ctx, test_prog_blocks)
    op = parser.parse_op()

    assert isinstance(op, TestOp)

    block1, block2, block3, block4 = op.regions[0].blocks

    assert block1.predecessors() == ()
    assert block2.predecessors() == (block1,)
    assert block3.predecessors() == (block1,)
    assert set(block4.predecessors()) == {block2, block3}



from collections.abc import Sequence

import pytest

from xdsl.dialects.builtin import Float64Type, MemRefType, f64
from xdsl.ir.affine import AffineMap
from xdsl.transforms.convert_memref_stream_to_snitch_stream import (
    strides_for_affine_map,
)
from xdsl.utils.exceptions import DiagnosticException


def mem_type(shape: Sequence[int]) -> MemRefType[Float64Type]:
    return MemRefType(f64, shape)


def test_strides_map_from_memref_type():
    with pytest.raises(
        DiagnosticException,
        match="Unsupported empty shape in memref of type memref<f64>",
    ):
        mem_type([]).get_affine_map_in_bytes()

    assert mem_type([2, 3]).get_affine_map_in_bytes() == AffineMap.from_callable(
        lambda i, j: (i * 24 + j * 8,)
    )


def test_strides_for_affine_map():
    assert strides_for_affine_map(AffineMap.identity(1), mem_type([2])) == [8]
    assert strides_for_affine_map(AffineMap.identity(2), mem_type([2, 3])) == [24, 8]
    assert strides_for_affine_map(AffineMap.identity(2), mem_type([3, 2])) == [16, 8]
    assert strides_for_affine_map(AffineMap.identity(3), mem_type([4, 3, 2])) == [
        48,
        16,
        8,
    ]
    assert strides_for_affine_map(AffineMap.transpose_map(), mem_type([3, 2])) == [
        8,
        16,
    ]
    assert strides_for_affine_map(AffineMap.transpose_map(), mem_type([2, 3])) == [
        8,
        24,
    ]



from dataclasses import dataclass

from xdsl.context import Context
from xdsl.dialects import memref_stream
from xdsl.dialects.builtin import ArrayAttr, IntAttr, ModuleOp
from xdsl.ir import SSAValue
from xdsl.passes import ModulePass
from xdsl.rewriter import Rewriter


def fold_fills_in_module(module_op: ModuleOp):
    fill_op_by_memref: dict[SSAValue, memref_stream.FillOp] = {}
    for op in module_op.walk():
        if isinstance(op, memref_stream.FillOp):
            if op.memref in fill_op_by_memref:
                # Two consecutive fills, erase first one, and replace with the new one
                Rewriter.erase_op(fill_op_by_memref[op.memref])
            fill_op_by_memref[op.memref] = op
            continue

        if isinstance(op, memref_stream.GenericOp):
            fill_ops = tuple(
                fill_op_by_memref.get(output, None) for output in op.outputs
            )
            indices = tuple(
                index for index, value in enumerate(fill_ops) if value is not None
            )
            if indices and op.is_imperfectly_nested:
                # There are values to rewrite, replace the operation
                init_indices = ArrayAttr(IntAttr(index) for index in indices)
                inits = tuple(
                    fill_op.value for fill_op in fill_ops if fill_op is not None
                )
                Rewriter.replace_op(
                    op,
                    memref_stream.GenericOp(
                        op.inputs,
                        op.outputs,
                        inits,
                        Rewriter.move_region_contents_to_new_regions(op.body),
                        op.indexing_maps,
                        op.iterator_types,
                        op.bounds,
                        init_indices,
                        op.doc,
                        op.library_call,
                    ),
                )
                for fill_op in set(value for value in fill_ops if value is not None):
                    Rewriter.erase_op(fill_op)

        for operand in op.operands:
            if operand in fill_op_by_memref:
                del fill_op_by_memref[operand]


@dataclass(frozen=True)
class MemRefStreamFoldFillPass(ModulePass):
    """
    Folds `memref_stream.fill` operations that run immediately before a
    `memref_stream.generic` operation into the init value.
    Assumes that none of the memrefs involved are aliased.
    """

    name = "memref-stream-fold-fill"

    def apply(self, ctx: Context, op: ModuleOp) -> None:
        fold_fills_in_module(op)



from xdsl.dialects.builtin import ModuleOp
from xdsl.dialects.test import TestOp
from xdsl.utils.op_selector import OpSelector


def test_op_selector():
    a = TestOp()
    b = TestOp()
    c = TestOp()

    module = ModuleOp([a, b, c])

    assert OpSelector(0, "builtin.module").get_op(module) is module
    assert OpSelector(1, "test.op").get_op(module) is a
    assert OpSelector(2, "test.op").get_op(module) is b
    assert OpSelector(3, "test.op").get_op(module) is c

    import pytest

    with pytest.raises(IndexError, match="Matching index 4 out of range."):
        OpSelector(4, "test.op").get_op(module)

    with pytest.raises(
        ValueError, match="Unexpected op builtin.module at index 0, expected test.op."
    ):
        OpSelector(0, "test.op").get_op(module)
    with pytest.raises(
        ValueError, match="Unexpected op test.op at index 1, expected builtin.module."
    ):
        OpSelector(1, "builtin.module").get_op(module)



from dataclasses import dataclass, field

from typing_extensions import TypeVar

from xdsl.context import Context
from xdsl.dialects.builtin import ModuleOp, UnregisteredOp
from xdsl.ir import Block, Operation, Region, Use
from xdsl.passes import ModulePass
from xdsl.pattern_rewriter import PatternRewriter
from xdsl.rewriter import Rewriter
from xdsl.traits import (
    IsolatedFromAbove,
    IsTerminator,
    MemoryEffectKind,
    get_effects,
    is_side_effect_free,
    only_has_effect,
)
from xdsl.transforms.dead_code_elimination import is_trivially_dead


@dataclass
class OperationInfo:
    """
    Boilerplate helper to use in KnownOps cache.

    This is to compare operations on their name, attributes, properties, results,
    operands, and matching region structure.
    """

    op: Operation

    @property
    def name(self):
        return (
            self.op.op_name.data
            if isinstance(self.op, UnregisteredOp)
            else self.op.name
        )

    def __hash__(self):
        return hash(
            (
                self.name,
                sum(hash(i) for i in self.op.attributes.items()),
                sum(hash(i) for i in self.op.properties.items()),
                hash(self.op.result_types),
                hash(self.op.operands),
            )
        )

    def __eq__(self, other: object):
        return (
            isinstance(other, OperationInfo)
            and hash(self) == hash(other)
            and self.name == other.name
            and self.op.attributes == other.op.attributes
            and self.op.properties == other.op.properties
            and self.op.operands == other.op.operands
            and self.op.result_types == other.op.result_types
            and all(
                s.is_structurally_equivalent(o)
                for s, o in zip(self.op.regions, other.op.regions, strict=True)
            )
        )


_D = TypeVar("_D")


class KnownOps:
    """
    Cache dictionary for known operations used in CSE.
    It quacks like a dict[Operation, Operation], but uses OperationInfo of an Operation
    as the actual key.
    """

    _known_ops: dict[OperationInfo, Operation]

    def __init__(self, known_ops: "KnownOps | None" = None):
        if known_ops is None:
            self._known_ops = {}
        else:
            self._known_ops = dict(known_ops._known_ops)

    def __getitem__(self, k: Operation):
        return self._known_ops[OperationInfo(k)]

    def __setitem__(self, k: Operation, v: Operation):
        self._known_ops[OperationInfo(k)] = v

    def __contains__(self, k: Operation):
        return OperationInfo(k) in self._known_ops

    def get(self, k: Operation, default: _D = None) -> Operation | _D:
        return self._known_ops.get(OperationInfo(k), default)

    def pop(self, k: Operation):
        return self._known_ops.pop(OperationInfo(k))


def has_other_side_effecting_op_in_between(
    from_op: Operation, to_op: Operation
) -> bool:
    """
    Returns if there *may* be a 'write' effecting operation between `from_op` and
    `to_op`.
    """
    assert from_op.parent is to_op.parent
    next_op = from_op
    while next_op is not to_op:
        effects = get_effects(next_op)
        if effects is None or any(e.kind is MemoryEffectKind.WRITE for e in effects):
            return True
        next_op = next_op.next_op
        assert next_op is not None, "Incorrect order of ops in side-effect search"
    return False


@dataclass
class CSEDriver:
    """
    Boilerplate class to handle and carry the state for CSE.
    """

    _rewriter: Rewriter | PatternRewriter = field(default_factory=Rewriter)
    _to_erase: set[Operation] = field(default_factory=set[Operation])
    _known_ops: KnownOps = field(default_factory=KnownOps)

    def _mark_erasure(self, op: Operation):
        self._to_erase.add(op)

    def _commit_erasures(self):
        for o in self._to_erase:
            if o.parent is not None:
                self._rewriter.erase_op(o)

    def _replace_and_delete(self, op: Operation, existing: Operation):
        """
        Factoring, replace `op` by `existing` and mark `op` for erasure.
        """

        # Just replace results
        def wasVisited(use: Use):
            return use.operation not in self._known_ops

        for o, n in zip(op.results, existing.results, strict=True):
            if all(wasVisited(u) for u in o.uses):
                o.replace_by(n)

        # If no uses remain, we can mark this operation for erasure
        if all(not r.uses for r in op.results):
            self._mark_erasure(op)

    def _simplify_operation(self, op: Operation):
        """
        Simplify a single operation: replace it by a corresponding known operation in
        scope, if any.
        Also just delete dead operations.
        """
        # Don't simplify terminators.
        if op.has_trait(IsTerminator):
            return

        # If the operation is already trivially dead just add it to the erase list.
        if is_trivially_dead(op):
            self._mark_erasure(op)
            return

        # Don't simplify operations with regions that have multiple blocks.
        # MLIR doesn't either at time of writing :)
        if any(len(region.blocks) > 1 for region in op.regions):
            return

        # Have a close look if the op might have side effects.
        if not is_side_effect_free(op):
            # If we can't be sure or the op has side effects, bail out
            if not only_has_effect(op, MemoryEffectKind.READ):
                return

            # If the op is only reading, we can still try to CSE it
            if existing := self._known_ops.get(op):
                if (
                    op.parent_block() is existing.parent_block()
                    # We then ensure there are no 'write' side-effecting operations
                    # in between the two, that could change the result of the operation
                    and not has_other_side_effecting_op_in_between(existing, op)
                ):
                    self._replace_and_delete(op, existing)
                    return

            # The operation is a CSE candidate, but we did not find a replacement
            # Mark it for any later occurence
            self._known_ops[op] = op
            return

        # If we know the operation is side-effect free, we can just replace it
        if existing := self._known_ops.get(op):
            self._replace_and_delete(op, existing)
            return

        # The operation is a CSE candidate, but we did not find a replacement
        # Mark it for any later occurence
        self._known_ops[op] = op

    def _simplify_block(self, block: Block):
        for op in block.ops:
            if op.regions:
                might_be_isolated = isinstance(op, UnregisteredOp) or (
                    op.get_trait(IsolatedFromAbove) is not None
                )
                # If we can't be sure the op isn't isolated, we assume it is for safety
                if might_be_isolated:
                    # Then save the current scope for later, but continue inside with a
                    # blank slate
                    old_scope = self._known_ops
                    self._known_ops = KnownOps()
                    for region in op.regions:
                        self._simplify_region(region)
                    self._known_ops = old_scope
                else:
                    for region in op.regions:
                        self._simplify_region(region)

            self._simplify_operation(op)

    def _simplify_region(self, region: Region):
        if not region.blocks:
            return

        if len(region.blocks) == 1:
            old_scope = self._known_ops
            self._known_ops = KnownOps(self._known_ops)

            self._simplify_block(region.block)

            self._known_ops = old_scope

    def simplify(self, thing: Operation | Block | Region):
        match thing:
            case Operation():
                for region in thing.regions:
                    self._simplify_region(region)
            case Block():
                self._simplify_block(thing)
            case Region():
                self._simplify_region(thing)
        self._commit_erasures()


def cse(
    thing: Operation | Block | Region,
    rewriter: Rewriter | PatternRewriter | None = None,
):
    if rewriter is not None:
        CSEDriver(_rewriter=rewriter).simplify(thing)
    else:
        CSEDriver().simplify(thing)


class CommonSubexpressionElimination(ModulePass):
    name = "cse"

    def apply(self, ctx: Context, op: ModuleOp) -> None:
        cse(op)



from xdsl.dialects.experimental import dmp
from xdsl.dialects.experimental.dmp import ExchangeDeclarationAttr, ShapeAttr


def flat_face_exchanges(
    shape: ShapeAttr, dim: int
) -> tuple[ExchangeDeclarationAttr, ...]:
    # we need access to the _flat_face_exchanges_for_dim method in order to test it
    # since this is a private function, and pyright will yell whenever it's accessed,
    # we have this wrapper function here that takes care of making the private publicly
    # accessible in the context of this test.
    func = dmp._flat_face_exchanges_for_dim  # pyright: ignore[reportPrivateUsage]
    return func(shape, dim)


def test_decomp_flat_face_3d():
    shape = ShapeAttr.from_index_attrs(
        (0, 0, 0),  # buff lb
        (4, 4, 4),  # core lb
        (14, 14, 14),  # core ub
        (18, 18, 18),  # buff ub
    )

    ex_pos_x, ex_neg_x = flat_face_exchanges(shape, 0)

    assert ex_pos_x.offset == (14, 4, 4)
    assert ex_pos_x.size == (4, 10, 10)
    assert ex_pos_x.source_offset == (-4, 0, 0)
    assert ex_pos_x.neighbor == (1, 0, 0)

    assert ex_neg_x.offset == (0, 4, 4)
    assert ex_neg_x.size == (4, 10, 10)
    assert ex_neg_x.source_offset == (4, 0, 0)
    assert ex_neg_x.neighbor == (-1, 0, 0)

    ex_pos_y, ex_neg_y = flat_face_exchanges(shape, 1)

    assert ex_pos_y.offset == (4, 14, 4)
    assert ex_pos_y.size == (10, 4, 10)
    assert ex_pos_y.source_offset == (0, -4, 0)
    assert ex_pos_y.neighbor == (0, 1, 0)

    assert ex_neg_y.offset == (4, 0, 4)
    assert ex_neg_y.size == (10, 4, 10)
    assert ex_neg_y.source_offset == (0, 4, 0)
    assert ex_neg_y.neighbor == (0, -1, 0)

    ex_pos_z, ex_neg_z = flat_face_exchanges(shape, 2)

    assert ex_pos_z.offset == (4, 4, 14)
    assert ex_pos_z.size == (10, 10, 4)
    assert ex_pos_z.source_offset == (0, 0, -4)
    assert ex_pos_z.neighbor == (0, 0, 1)

    assert ex_neg_z.offset == (4, 4, 0)
    assert ex_neg_z.size == (10, 10, 4)
    assert ex_neg_z.source_offset == (0, 0, 4)
    assert ex_neg_z.neighbor == (0, 0, -1)


def test_decomp_flat_face_4d():
    shape = ShapeAttr.from_index_attrs(
        (0, 0, 0, 0),  # buff lb
        (4, 4, 4, 4),  # core lb
        (14, 14, 14, 14),  # core ub
        (18, 18, 18, 18),  # buff ub
    )

    ex_pos_y, ex_neg_y = flat_face_exchanges(shape, 1)

    assert ex_pos_y.offset == (4, 14, 4, 4)
    assert ex_pos_y.size == (10, 4, 10, 10)
    assert ex_pos_y.source_offset == (0, -4, 0, 0)
    assert ex_pos_y.neighbor == (0, 1, 0, 0)

    assert ex_neg_y.offset == (4, 0, 4, 4)
    assert ex_neg_y.size == (10, 4, 10, 10)
    assert ex_neg_y.source_offset == (0, 4, 0, 0)
    assert ex_neg_y.neighbor == (0, -1, 0, 0)


def test_decomp_with_overflow():
    shape = ShapeAttr.from_index_attrs(
        (0, 0, 0),  # buff lb
        (2, 2, 2),  # core lb
        (3, 3, 102),  # core ub
        (5, 5, 104),  # buff ub
    )

    exchanges = tuple(flat_face_exchanges(shape, 0))
    assert len(exchanges) == 4

    ex_px_1, ex_px_2, ex_nx_1, ex_nx_2 = exchanges

    # all exchanges are of size (1, 1, 100)
    assert all(ex.size == (1, 1, 100) for ex in exchanges)

    # the first exchange is closer to the core region
    assert ex_px_1.offset == (3, 2, 2)
    # and has a source offset of (-1, 0, 0)
    assert ex_px_1.source_offset == (-1, 0, 0)
    # the second exchange is farther away
    assert ex_px_2.offset == (4, 2, 2)
    # and has a source offset of twice that
    assert ex_px_2.source_offset == (-2, 0, 0)

    # same for negative x, first exchange is closer to the core
    assert ex_nx_1.offset == (1, 2, 2)
    # and has a source offset of (1, 0, 0)
    assert ex_nx_1.source_offset == (1, 0, 0)
    # second is farther away
    assert ex_nx_2.offset == (0, 2, 2)
    # and has a source offset of (2, 0, 0)
    assert ex_nx_2.source_offset == (2, 0, 0)

    assert ex_px_1.neighbor == (1, 0, 0)
    assert ex_px_2.neighbor == (2, 0, 0)

    assert ex_nx_1.neighbor == (-1, 0, 0)
    assert ex_nx_2.neighbor == (-2, 0, 0)



from collections.abc import Callable
from io import StringIO

import pytest

from xdsl.context import Context
from xdsl.dialects import test
from xdsl.dialects.arith import AddiOp, Arith, ConstantOp
from xdsl.dialects.builtin import Builtin, Float32Type, Float64Type, ModuleOp, i32, i64
from xdsl.ir import Block, Region
from xdsl.parser import Parser
from xdsl.printer import Printer
from xdsl.rewriter import BlockInsertPoint, InsertPoint, Rewriter


def rewrite_and_compare(
    prog: str, expected_prog: str, transformation: Callable[[ModuleOp, Rewriter], None]
):
    ctx = Context()
    ctx.load_dialect(Builtin)
    ctx.load_dialect(Arith)
    ctx.load_dialect(test.Test)

    parser = Parser(ctx, prog)
    module = parser.parse_module()

    rewriter = Rewriter()
    transformation(module, rewriter)

    file = StringIO()
    printer = Printer(stream=file, print_generic_format=True)
    printer.print_op(module)

    assert file.getvalue().strip() == expected_prog.strip()


def test_operation_deletion():
    """Test rewrites where SSA values are deleted."""

    prog = """\
"builtin.module"() ({
  %0 = "arith.constant"() <{value = 5 : i32}> : () -> i32
}) : () -> ()
"""

    expected = """\
"builtin.module"() ({
^0:
}) : () -> ()"""

    def transformation(module: ModuleOp, rewriter: Rewriter) -> None:
        constant_op = module.ops.first
        assert constant_op is not None
        rewriter.erase_op(constant_op)

    rewrite_and_compare(prog, expected, transformation)


# Test an operation replacement
def test_replace_op_one_op():
    prog = """\
"builtin.module"() ({
  %0 = "arith.constant"() <{value = 42 : i32}> : () -> i32
  %1 = "arith.addi"(%0, %0) <{overflowFlags = #arith.overflow<none>}> : (i32, i32) -> i32
}) : () -> ()
"""

    expected = """\
"builtin.module"() ({
  %0 = "arith.constant"() <{value = 43 : i32}> : () -> i32
  %1 = "arith.addi"(%0, %0) <{overflowFlags = #arith.overflow<none>}> : (i32, i32) -> i32
}) : () -> ()
"""

    def transformation(module: ModuleOp, rewriter: Rewriter) -> None:
        constant_op = module.ops.first
        assert constant_op is not None
        new_constant_op = ConstantOp.from_int_and_width(43, i32)
        rewriter.replace_op(constant_op, new_constant_op)

    rewrite_and_compare(prog, expected, transformation)


# Test an operation replacement with multiple ops
def test_replace_op_multiple_op():
    prog = """\
"builtin.module"() ({
  %0 = "arith.constant"() <{value = 2 : i32}> : () -> i32
  %1 = "arith.addi"(%0, %0) <{overflowFlags = #arith.overflow<none>}> : (i32, i32) -> i32
}) : () -> ()
"""

    expected = """\
"builtin.module"() ({
  %0 = "arith.constant"() <{value = 1 : i32}> : () -> i32
  %1 = "arith.addi"(%0, %0) <{overflowFlags = #arith.overflow<none>}> : (i32, i32) -> i32
  %2 = "arith.addi"(%1, %1) <{overflowFlags = #arith.overflow<none>}> : (i32, i32) -> i32
}) : () -> ()
"""

    def transformation(module: ModuleOp, rewriter: Rewriter) -> None:
        constant_op = module.ops.first
        assert constant_op is not None
        new_constant = ConstantOp.from_int_and_width(1, i32)
        new_add = AddiOp(new_constant, new_constant)

        rewriter.replace_op(constant_op, [new_constant, new_add])

    rewrite_and_compare(prog, expected, transformation)


# Test an operation replacement with manually specified results
def test_replace_op_new_results():
    prog = """\
"builtin.module"() ({
  %0 = "arith.constant"() <{value = 2 : i32}> : () -> i32
  %1 = "arith.addi"(%0, %0) <{overflowFlags = #arith.overflow<none>}> : (i32, i32) -> i32
  %2 = "arith.muli"(%1, %1) <{overflowFlags = #arith.overflow<none>}> : (i32, i32) -> i32
}) : () -> ()
"""

    expected = """\
"builtin.module"() ({
  %0 = "arith.constant"() <{value = 2 : i32}> : () -> i32
  %1 = "arith.muli"(%0, %0) <{overflowFlags = #arith.overflow<none>}> : (i32, i32) -> i32
}) : () -> ()
"""

    def transformation(module: ModuleOp, rewriter: Rewriter) -> None:
        ops_iter = iter(module.ops)
        next(ops_iter)
        add_op = next(ops_iter)
        assert isinstance(add_op, AddiOp)

        rewriter.replace_op(add_op, [], [add_op.lhs])

    rewrite_and_compare(prog, expected, transformation)


def test_inline_block_at_end():
    """Test the inlining of a block at end."""

    prog = """\
"builtin.module"() ({
  %0 = "test.op"() : () -> !test.type<"int">
  "test.op"() ({
    %1 = "test.op"() : () -> !test.type<"int">
  }) : () -> ()
}) : () -> ()
"""

    expected = """\
"builtin.module"() ({
  %0 = "test.op"() : () -> !test.type<"int">
  "test.op"() ({
  }) : () -> ()
  %1 = "test.op"() : () -> !test.type<"int">
}) : () -> ()
"""

    def transformation(module: ModuleOp, rewriter: Rewriter) -> None:
        ops_iter = iter(module.ops)
        next(ops_iter)
        test_op = next(ops_iter)
        module_block = module.regions[0].blocks[0]
        test_block = test_op.regions[0].blocks[0]

        rewriter.inline_block(test_block, InsertPoint.at_end(module_block))

    rewrite_and_compare(prog, expected, transformation)


def test_inline_block_before():
    """Test the inlining of a block before an operation."""

    prog = """\
"builtin.module"() ({
  %0 = "test.op"() : () -> !test.type<"int">
  "test.op"() ({
    %1 = "test.op"() : () -> !test.type<"int">
  }) : () -> ()
}) : () -> ()
"""

    expected = """\
"builtin.module"() ({
  %0 = "test.op"() : () -> !test.type<"int">
  %1 = "test.op"() : () -> !test.type<"int">
  "test.op"() ({
  }) : () -> ()
}) : () -> ()
"""

    def transformation(module: ModuleOp, rewriter: Rewriter) -> None:
        ops_iter = iter(module.ops)
        next(ops_iter)
        test_op = next(ops_iter)
        test_block = test_op.regions[0].blocks[0]

        rewriter.inline_block(test_block, InsertPoint.before(test_op))

    rewrite_and_compare(prog, expected, transformation)


def test_inline_block_after():
    """Test the inlining of a block after an operation."""

    prog = """\
"builtin.module"() ({
  %0 = "test.op"() : () -> !test.type<"int">
  "test.op"() ({
    %1 = "test.op"() : () -> !test.type<"int">
  }) : () -> ()
}) : () -> ()
"""

    expected = """\
"builtin.module"() ({
  %0 = "test.op"() : () -> !test.type<"int">
  %1 = "test.op"() : () -> !test.type<"int">
  "test.op"() ({
  }) : () -> ()
}) : () -> ()
"""

    def transformation(module: ModuleOp, rewriter: Rewriter) -> None:
        ops_iter = iter(module.ops)
        constant_op = next(ops_iter)
        test_op = next(ops_iter)
        test_block = test_op.regions[0].blocks[0]

        rewriter.inline_block(test_block, InsertPoint.after(constant_op))

    rewrite_and_compare(prog, expected, transformation)


def test_insert_block():
    """Test the insertion of a block in a region."""
    prog = """\
"builtin.module"() ({
  %0 = "arith.constant"() <{value = true}> : () -> i1
}) : () -> ()
"""

    expected = """\
"builtin.module"() ({
^0:
^1:
  %0 = "arith.constant"() <{value = true}> : () -> i1
}) : () -> ()
"""

    def transformation(module: ModuleOp, rewriter: Rewriter) -> None:
        module.regions[0].insert_block(Block(), 0)

    rewrite_and_compare(prog, expected, transformation)


def test_insert_block2():
    """Test the insertion of a block in a region."""
    prog = """\
"builtin.module"() ({
  %0 = "arith.constant"() <{value = true}> : () -> i1
}) : () -> ()
"""

    expected = """\
"builtin.module"() ({
  %0 = "arith.constant"() <{value = true}> : () -> i1
^0:
}) : () -> ()
"""

    def transformation(module: ModuleOp, rewriter: Rewriter) -> None:
        module.regions[0].insert_block(Block(), 1)

    rewrite_and_compare(prog, expected, transformation)


def test_insert_block_before():
    """Test the insertion of a block before another block."""
    prog = """\
"builtin.module"() ({
  %0 = "arith.constant"() <{value = true}> : () -> i1
}) : () -> ()
"""

    expected = """\
"builtin.module"() ({
^0:
^1:
  %0 = "arith.constant"() <{value = true}> : () -> i1
}) : () -> ()
"""

    def insert_empty_block_before(module: ModuleOp, rewriter: Rewriter) -> None:
        rewriter.insert_block(
            Block(), BlockInsertPoint.before(module.regions[0].blocks[0])
        )

    rewrite_and_compare(prog, expected, insert_empty_block_before)


def test_insert_block_after():
    """Test the insertion of a block after another block."""
    prog = """\
"builtin.module"() ({
  %0 = "arith.constant"() <{value = true}> : () -> i1
}) : () -> ()


"""

    expected = """\
"builtin.module"() ({
  %0 = "arith.constant"() <{value = true}> : () -> i1
^0:
}) : () -> ()
"""

    def insert_empty_block_after(module: ModuleOp, rewriter: Rewriter) -> None:
        rewriter.insert_block(
            Block(), BlockInsertPoint.after(module.regions[0].blocks[0])
        )

    rewrite_and_compare(prog, expected, insert_empty_block_after)


def test_insert_op_before():
    """Test the insertion of an operation before another operation."""
    prog = """\
"builtin.module"() ({
  %0 = "arith.constant"() <{value = 43 : i32}> : () -> i32
}) : () -> ()
"""

    expected = """\
"builtin.module"() ({
  %0 = "arith.constant"() <{value = 34 : i64}> : () -> i64
  %1 = "arith.constant"() <{value = 43 : i32}> : () -> i32
}) : () -> ()
"""

    def transformation(module: ModuleOp, rewriter: Rewriter) -> None:
        constant = ConstantOp.from_int_and_width(34, i64)
        first_op = module.regions[0].blocks[0].first_op
        assert first_op is not None
        rewriter.insert_op(constant, InsertPoint.before(first_op))

    rewrite_and_compare(prog, expected, transformation)


def test_insert_op_after():
    """Test the insertion of an operation after another operation."""
    prog = """\
"builtin.module"() ({
  %0 = "arith.constant"() <{value = 43 : i32}> : () -> i32
}) : () -> ()
"""

    expected = """\
"builtin.module"() ({
  %0 = "arith.constant"() <{value = 43 : i32}> : () -> i32
  %1 = "arith.constant"() <{value = 34 : i64}> : () -> i64
}) : () -> ()
"""

    def transformation(module: ModuleOp, rewriter: Rewriter) -> None:
        constant = ConstantOp.from_int_and_width(34, i64)
        first_op = module.regions[0].blocks[0].first_op
        assert first_op is not None
        rewriter.insert_op(constant, InsertPoint.after(first_op))

    rewrite_and_compare(prog, expected, transformation)


def test_preserve_naming_single_op():
    """Test the preservation of names of SSAValues"""
    prog = """\
"builtin.module"() ({
  %i = "arith.constant"() <{value = 42 : i32}> : () -> i32
  %1 = "arith.addi"(%i, %i) <{overflowFlags = #arith.overflow<none>}> : (i32, i32) -> i32
}) : () -> ()
"""

    expected = """\
"builtin.module"() ({
  %i = "arith.constant"() <{value = 1 : i32}> : () -> i32
  %0 = "arith.addi"(%i, %i) <{overflowFlags = #arith.overflow<none>}> : (i32, i32) -> i32
}) : () -> ()
"""

    def transformation(module: ModuleOp, rewriter: Rewriter) -> None:
        constant_op = module.ops.first
        assert constant_op is not None
        new_constant = ConstantOp.from_int_and_width(1, i32)

        rewriter.replace_op(constant_op, [new_constant])

    rewrite_and_compare(prog, expected, transformation)


def test_preserve_naming_multiple_ops():
    """Test the preservation of names of SSAValues for transformations to multiple ops"""
    prog = """\
"builtin.module"() ({
  %i = "arith.constant"() <{value = 42 : i32}> : () -> i32
  %1 = "arith.addi"(%i, %i) <{overflowFlags = #arith.overflow<none>}> : (i32, i32) -> i32
}) : () -> ()
"""

    expected = """\
"builtin.module"() ({
  %i = "arith.constant"() <{value = 1 : i32}> : () -> i32
  %i_1 = "arith.addi"(%i, %i) <{overflowFlags = #arith.overflow<none>}> : (i32, i32) -> i32
  %0 = "arith.addi"(%i_1, %i_1) <{overflowFlags = #arith.overflow<none>}> : (i32, i32) -> i32
}) : () -> ()
"""

    def transformation(module: ModuleOp, rewriter: Rewriter) -> None:
        constant_op = module.ops.first
        assert constant_op is not None
        new_constant = ConstantOp.from_int_and_width(1, i32)
        new_add = AddiOp(new_constant, new_constant)

        rewriter.replace_op(constant_op, [new_constant, new_add])

    rewrite_and_compare(prog, expected, transformation)


def test_no_result_rewriter():
    """Test rewriter on ops without results"""
    prog = """\
"builtin.module"() ({
  "test.termop"() : () -> ()
}) : () -> ()
"""

    expected = """\
"builtin.module"() ({
  "test.op"() : () -> ()
}) : () -> ()
"""

    def transformation(module: ModuleOp, rewriter: Rewriter) -> None:
        return_op = module.ops.first
        assert return_op is not None
        new_op = test.TestOp.create()

        rewriter.replace_op(return_op, [new_op])

    rewrite_and_compare(prog, expected, transformation)


# Test erase operation
def test_erase_op():
    prog = """\
"builtin.module"() ({
  %0 = "arith.constant"() <{value = 42 : i32}> : () -> i32
  %1 = "arith.addi"(%0, %0) <{overflowFlags = #arith.overflow<none>}> : (i32, i32) -> i32
}) : () -> ()
"""

    expected = """\
"builtin.module"() ({
  %0 = "arith.addi"(%1, %1) <{overflowFlags = #arith.overflow<none>}> : (i32, i32) -> i32
}) : () -> ()
"""

    def transformation_safe(module: ModuleOp, rewriter: Rewriter) -> None:
        constant_op = module.ops.first
        assert constant_op is not None
        rewriter.erase_op(constant_op, safe_erase=True)

    def transformation_unsafe(module: ModuleOp, rewriter: Rewriter) -> None:
        constant_op = module.ops.first
        assert constant_op is not None
        rewriter.erase_op(constant_op, safe_erase=False)

    rewrite_and_compare(prog, expected, transformation_unsafe)

    with pytest.raises(
        Exception,
        match="Attempting to delete SSA value that still has uses",
    ):
        rewrite_and_compare(prog, expected, transformation_safe)


def test_erase_orphan_op():
    """Test that we can erase an orphan operation."""
    module = ModuleOp([])
    rewriter = Rewriter()
    rewriter.erase_op(module)


def test_inline_region_before():
    """Test the insertion of a block in a region."""
    prog = """\
"builtin.module"() ({
  %0 = "test.op"() : () -> i32
^0:
  %1 = "test.op"() : () -> i64
}) : () -> ()
"""

    expected = """\
"builtin.module"() ({
  %0 = "test.op"() : () -> i32
^0:
  %1 = "test.op"() : () -> f32
^1:
  %2 = "test.op"() : () -> f64
^2:
  %3 = "test.op"() : () -> i64
}) : () -> ()
"""

    def transformation(module: ModuleOp, rewriter: Rewriter) -> None:
        region = Region(
            (
                Block((test.TestOp(result_types=(Float32Type(),)),)),
                Block((test.TestOp(result_types=(Float64Type(),)),)),
            )
        )
        rewriter.inline_region(region, BlockInsertPoint.before(module.body.blocks[1]))

    rewrite_and_compare(prog, expected, transformation)


def test_inline_region_after():
    """Test the insertion of a block in a region."""
    prog = """\
"builtin.module"() ({
  %0 = "test.op"() : () -> i32
^0:
  %1 = "test.op"() : () -> i64
}) : () -> ()
"""

    expected = """\
"builtin.module"() ({
  %0 = "test.op"() : () -> i32
^0:
  %1 = "test.op"() : () -> f32
^1:
  %2 = "test.op"() : () -> f64
^2:
  %3 = "test.op"() : () -> i64
}) : () -> ()
"""

    def transformation(module: ModuleOp, rewriter: Rewriter) -> None:
        region = Region(
            (
                Block((test.TestOp(result_types=(Float32Type(),)),)),
                Block((test.TestOp(result_types=(Float64Type(),)),)),
            )
        )
        rewriter.inline_region(region, BlockInsertPoint.after(module.body.blocks[0]))

    rewrite_and_compare(prog, expected, transformation)


def test_inline_region_at_start():
    """Test the insertion of a block in a region."""
    prog = """\
"builtin.module"() ({
  %0 = "test.op"() : () -> i32
^0:
  %1 = "test.op"() : () -> i64
}) : () -> ()
"""

    expected = """\
"builtin.module"() ({
  %0 = "test.op"() : () -> f32
^0:
  %1 = "test.op"() : () -> f64
^1:
  %2 = "test.op"() : () -> i32
^2:
  %3 = "test.op"() : () -> i64
}) : () -> ()
"""

    def transformation(module: ModuleOp, rewriter: Rewriter) -> None:
        region = Region(
            (
                Block((test.TestOp(result_types=(Float32Type(),)),)),
                Block((test.TestOp(result_types=(Float64Type(),)),)),
            )
        )
        rewriter.inline_region(region, BlockInsertPoint.at_start(module.body))

    rewrite_and_compare(prog, expected, transformation)


def test_inline_region_at_end():
    """Test the insertion of a block in a region."""
    prog = """\
"builtin.module"() ({
  %0 = "test.op"() : () -> i32
^0:
  %1 = "test.op"() : () -> i64
}) : () -> ()
"""

    expected = """\
"builtin.module"() ({
  %0 = "test.op"() : () -> i32
^0:
  %1 = "test.op"() : () -> i64
^1:
  %2 = "test.op"() : () -> f32
^2:
  %3 = "test.op"() : () -> f64
}) : () -> ()
"""

    def transformation(module: ModuleOp, rewriter: Rewriter) -> None:
        region = Region(
            (
                Block((test.TestOp(result_types=(Float32Type(),)),)),
                Block((test.TestOp(result_types=(Float64Type(),)),)),
            )
        )
        rewriter.inline_region(region, BlockInsertPoint.at_end(module.body))

    rewrite_and_compare(prog, expected, transformation)


def test_verify_inline_region():
    region = Region(Block())

    with pytest.raises(ValueError, match="Cannot move region into itself."):
        Rewriter.inline_region(region, BlockInsertPoint.before(region.block))

    with pytest.raises(ValueError, match="Cannot move region into itself."):
        Rewriter.inline_region(region, BlockInsertPoint.after(region.block))

    with pytest.raises(ValueError, match="Cannot move region into itself."):
        Rewriter.inline_region(region, BlockInsertPoint.at_start(region))

    with pytest.raises(ValueError, match="Cannot move region into itself."):
        Rewriter.inline_region(region, BlockInsertPoint.at_end(region))



from __future__ import annotations

from typing import ClassVar

from typing_extensions import Self

from xdsl.dialects.builtin import (
    BFloat16Type,
    ContainerOf,
    DenseIntElementsAttr,
    Float16Type,
    Float32Type,
    Float64Type,
    Float80Type,
    Float128Type,
    IndexType,
    IntegerType,
    VectorType,
)
from xdsl.ir import Attribute, Dialect, Operation, SSAValue
from xdsl.irdl import (
    AnyAttr,
    AnyOf,
    IRDLOperation,
    VarConstraint,
    irdl_op_definition,
    operand_def,
    prop_def,
    result_def,
    traits_def,
    var_operand_def,
)
from xdsl.parser import Parser
from xdsl.printer import Printer
from xdsl.traits import Pure

integerOrFloatLike = ContainerOf(
    AnyOf(
        [
            IntegerType,
            IndexType,
            BFloat16Type,
            Float16Type,
            Float32Type,
            Float64Type,
            Float80Type,
            Float128Type,
        ]
    )
)


class VarithOp(IRDLOperation):
    """
    Variadic arithmetic operation
    """

    T: ClassVar = VarConstraint("T", integerOrFloatLike)

    args = var_operand_def(T)
    res = result_def(T)

    traits = traits_def(Pure())

    assembly_format = "$args attr-dict `:` type($res)"

    def __init__(self, *args: SSAValue | Operation):
        assert len(args) > 0
        super().__init__(operands=[args], result_types=[SSAValue.get(args[-1]).type])


@irdl_op_definition
class VarithAddOp(VarithOp):
    name = "varith.add"


@irdl_op_definition
class VarithMulOp(VarithOp):
    name = "varith.mul"


@irdl_op_definition
class VarithSwitchOp(IRDLOperation):
    """
    Variadic selection operation

    Similar to `cf.switch`, this operation returns the argument corresponding to
    `flag`, returning the default value otherwise.
    """

    name = "varith.switch"

    T: ClassVar = VarConstraint("T", AnyAttr())

    flag = operand_def(IntegerType | IndexType)
    case_values = prop_def(DenseIntElementsAttr)

    default_arg = operand_def(T)
    args = var_operand_def(T)

    result = result_def(T)

    traits = traits_def(Pure())

    def __init__(
        self,
        flag: SSAValue | Operation,
        case_values: DenseIntElementsAttr,
        default_arg: SSAValue | Operation,
        *args: SSAValue | Operation,
        attr_dict: dict[str, Attribute] | None = None,
    ):
        super().__init__(
            operands=[
                flag,
                default_arg,
                args,
            ],
            properties={
                "case_values": case_values,
            },
            attributes=attr_dict,
            result_types=(SSAValue.get(default_arg).type,),
        )

    @classmethod
    def parse(cls, parser: Parser) -> Self:
        unresolved_flag = parser.parse_unresolved_operand()
        parser.parse_punctuation(":")
        flag_type = parser.parse_type()
        assert isinstance(flag_type, IntegerType | IndexType)
        flag = parser.resolve_operand(unresolved_flag, flag_type)
        parser.parse_punctuation("->")
        return_type = parser.parse_type()
        parser.parse_punctuation(",")
        parser.parse_punctuation("[")
        parser.parse_keyword("default")
        parser.parse_punctuation(":")
        default_arg = parser.resolve_operand(
            parser.parse_unresolved_operand(), return_type
        )

        values: list[int] = []
        args: list[SSAValue] = []
        while parser.parse_optional_punctuation(","):
            values.append(parser.parse_integer())
            parser.parse_punctuation(":")
            args.append(
                parser.resolve_operand(parser.parse_unresolved_operand(), return_type)
            )
        parser.parse_punctuation("]")
        attr_dict = parser.parse_optional_attr_dict()

        case_values = DenseIntElementsAttr.from_list(
            VectorType(flag_type, (len(values),)), values
        )

        return cls(
            flag,
            case_values,
            default_arg,
            *args,
            attr_dict=attr_dict,
        )

    @staticmethod
    def _print_case(printer: Printer, case_name: str, arg: SSAValue):
        printer.print_string(case_name)
        printer.print_string(": ")
        printer.print_operand(arg)

    def print(self, printer: Printer):
        printer.print_string(" ")
        printer.print_operand(self.flag)
        printer.print_string(" : ")
        printer.print_attribute(self.flag.type)
        printer.print_string(" -> ")
        printer.print_attribute(self.result.type)
        printer.print_string(", [")
        with printer.indented():
            printer.print_string("\n")
            cases = [("default", self.default_arg)] + [
                (str(c), arg)
                for (c, arg) in zip(
                    self.case_values.get_values(),
                    self.args,
                    strict=True,
                )
            ]
            printer.print_list(
                cases, lambda x: self._print_case(printer, x[0], x[1]), ",\n"
            )
        printer.print_string("\n]")
        if self.attributes:
            printer.print_attr_dict(self.attributes)


Varith = Dialect(
    "varith",
    [
        VarithAddOp,
        VarithMulOp,
        VarithSwitchOp,
    ],
)



from __future__ import annotations

from collections.abc import Iterable
from dataclasses import dataclass

from xdsl.builder import Builder, InsertPoint
from xdsl.dialects.builtin import (
    FileLineColLoc,
    IntAttr,
    ModuleOp,
    StringAttr,
    TensorType,
    UnrankedTensorType,
    f64,
)
from xdsl.ir import Block, Region, SSAValue
from xdsl.utils.lexer import Location
from xdsl.utils.scoped_dict import ScopedDict

from ..dialects.toy import (
    AddOp,
    AnyTensorTypeF64,
    ConstantOp,
    FuncOp,
    FunctionType,
    GenericCallOp,
    MulOp,
    PrintOp,
    ReshapeOp,
    ReturnOp,
    TransposeOp,
    UnrankedTensorTypeF64,
)
from .toy_ast import (
    BinaryExprAST,
    CallExprAST,
    ExprAST,
    FunctionAST,
    LiteralExprAST,
    ModuleAST,
    NumberExprAST,
    PrintExprAST,
    PrototypeAST,
    ReturnExprAST,
    VarDeclExprAST,
    VariableExprAST,
)


class IRGenError(Exception):
    pass


@dataclass(init=False)
class IRGen:
    """
    Implementation of a simple MLIR emission from the Toy AST.

    This will emit operations that are specific to the Toy language, preserving
    the semantics of the language and (hopefully) allow to perform accurate
    analysis and transformation based on these high level semantics.
    """

    module: ModuleOp
    """A "module" matches a Toy source file: containing a list of functions."""

    builder: Builder
    """
    The builder is a helper class to create IR inside a function. The builder
    is stateful, in particular it keeps an "insertion point": this is where
    the next operations will be introduced."""

    symbol_table: ScopedDict[str, SSAValue] | None = None
    """
    The symbol table maps a variable name to a value in the current scope.
    Entering a function creates a new scope, and the function arguments are
    added to the mapping. When the processing of a function is terminated, the
    scope is destroyed and the mappings created in this scope are dropped."""

    def __init__(self):
        # We create an empty MLIR module and codegen functions one at a time and
        # add them to the module.
        self.module = ModuleOp([])
        self.builder = Builder(InsertPoint.at_end(self.module.body.blocks[0]))

    def ir_gen_module(self, module_ast: ModuleAST) -> ModuleOp:
        """
        Public API: convert the AST for a Toy module (source file) to an MLIR
        Module operation."""

        for f in module_ast.funcs:
            self.ir_gen_function(f)

        # Verify the module after we have finished constructing it, this will check
        # the structural properties of the IR and invoke any specific verifiers we
        # have on the Toy operations.
        try:
            self.module.verify()
        except Exception:
            print("module verification error")
            raise

        return self.module

    def loc(self, loc: Location):
        "Helper conversion for a Toy AST location to an MLIR location."
        return FileLineColLoc(StringAttr(loc.file), IntAttr(loc.line), IntAttr(loc.col))

    def declare(self, var: str, value: SSAValue) -> bool:
        """
        Declare a variable in the current scope, return success if the variable
        wasn't declared yet."""
        assert self.symbol_table is not None
        if var in self.symbol_table:
            return False
        self.symbol_table[var] = value
        return True

    def get_type(self, shape: list[int]) -> AnyTensorTypeF64:
        "Build a tensor type from a list of shape dimensions."
        # If the shape is empty, then this type is unranked.
        if len(shape):
            return TensorType(f64, shape)
        else:
            return UnrankedTensorTypeF64(f64)

    def ir_gen_proto(self, proto_ast: PrototypeAST) -> FuncOp:
        """
        Create the prototype for a function with as many arguments as the
        provided Toy AST prototype."""
        # location = self.loc(proto_ast.loc)

        # This is a generic function, the return type will be inferred later.
        # Arguments type are uniformly unranked tensors.
        func_type = FunctionType.from_lists(
            [self.get_type([])] * len(proto_ast.args), [self.get_type([])]
        )
        return self.builder.insert(FuncOp(proto_ast.name, func_type, Region()))

    def ir_gen_function(self, function_ast: FunctionAST) -> FuncOp:
        "Emit a new function and add it to the MLIR module."

        # keep builder for later
        parent_builder = self.builder

        # Create a scope in the symbol table to hold variable declarations.
        self.symbol_table = ScopedDict()

        proto_args = function_ast.proto.args

        # Create the block for the current function
        block = Block(
            arg_types=[UnrankedTensorType(f64) for _ in range(len(proto_args))]
        )
        self.builder = Builder(InsertPoint.at_end(block))

        # Declare all the function arguments in the symbol table.
        for name, value in zip(proto_args, block.args):
            self.declare(name.name, value)

        # Emit the body of the function.
        self.ir_gen_expr_list(function_ast.body)

        return_types = []

        # Implicitly return void if no return statement was emitted.
        return_op = None
        if block.ops:
            last_op = block.last_op
            if isinstance(last_op, ReturnOp):
                return_op = last_op
                if return_op.input is not None:
                    return_arg = return_op.input
                    return_types = [return_arg.type]
        if return_op is None:
            self.builder.insert(ReturnOp())

        input_types = [self.get_type([]) for _ in range(len(function_ast.proto.args))]

        func_type = FunctionType.from_lists(input_types, return_types)

        # main should be public, all the others private
        private = function_ast.proto.name != "main"

        # clean up
        self.symbol_table = None
        self.builder = parent_builder

        func = self.builder.insert(
            FuncOp(function_ast.proto.name, func_type, Region(block), private=private)
        )

        return func

    def ir_gen_binary_expr(self, binop: BinaryExprAST) -> SSAValue:
        "Emit a binary operation"

        # First emit the operations for each side of the operation before emitting
        # the operation itself. For example if the expression is `a + foo(a)`
        # 1) First it will visiting the LHS, which will return a reference to the
        #    value holding `a`. This value should have been emitted at declaration
        #    time and registered in the symbol table, so nothing would be
        #    codegen'd. If the value is not in the symbol table, an error has been
        #    emitted and nullptr is returned.
        # 2) Then the RHS is visited (recursively) and a call to `foo` is emitted
        #    and the result value is returned. If an error occurs we get a nullptr
        #    and propagate.

        lhs = self.ir_gen_expr(binop.lhs)
        rhs = self.ir_gen_expr(binop.rhs)

        # location = self.loc(binop.loc)

        # Derive the operation name from the binary operator. At the moment we only
        # support '+' and '*'.
        if binop.op == "+":
            op = self.builder.insert(AddOp(lhs, rhs))
        elif binop.op == "*":
            op = self.builder.insert(MulOp(lhs, rhs))
        else:
            raise IRGenError(f"Unsupported binary operation `{binop.op}`")

        return op.res

    def ir_gen_variable_expr(self, expr: VariableExprAST) -> SSAValue:
        """
        This is a reference to a variable in an expression. The variable is
        expected to have been declared and so should have a value in the symbol
        table, otherwise emit an error and return nullptr."""
        assert self.symbol_table is not None
        try:
            variable = self.symbol_table[expr.name]
            return variable
        except Exception as e:
            raise IRGenError(f"error: unknown variable `{expr.name}`") from e

    def ir_gen_return_expr(self, ret: ReturnExprAST):
        "Emit a return operation. This will return failure if any generation fails."

        # location = self.loc(binop.loc)

        # 'return' takes an optional expression, handle that case here.
        if ret.expr is not None:
            expr = self.ir_gen_expr(ret.expr)
        else:
            expr = None

        self.builder.insert(ReturnOp(expr))

    def ir_gen_literal_expr(self, lit: LiteralExprAST) -> SSAValue:
        """
        Emit a literal/constant array. It will be emitted as a flattened array of
        data in an Attribute attached to a `toy.constant` operation.
        See documentation on [Attributes](LangRef.md#attributes) for more details.
        Here is an excerpt:
        Attributes are the mechanism for specifying constant data in MLIR in
        places where a variable is never allowed [...]. They consist of a name
        and a concrete attribute value. The set of expected attributes, their
        structure, and their interpretation are all contextually dependent on
        what they are attached to.
        Example, the source level statement:
        var a<2, 3> = [[1, 2, 3], [4, 5, 6]];
        will be converted to:
        %0 = "toy.constant"() {value: dense<tensor<2x3xf64>,
            [[1.000000e+00, 2.000000e+00, 3.000000e+00],
            [4.000000e+00, 5.000000e+00, 6.000000e+00]]>} : () -> tensor<2x3xf64>
        """

        # The attribute is a vector with a integer value per element
        # (number) in the array, see `collectData()` below for more details.
        data = self.collect_data(lit)

        # Build the MLIR op `toy.constant`. This invokes the `ConstantOp::build`
        # method.
        op = self.builder.insert(ConstantOp.from_list(data, lit.dims))
        return op.res

    def collect_data(self, expr: ExprAST) -> list[float]:
        """
        Helper function to accumulate the data that compose an array
        literal. It flattens the nested structure in the supplied vector. For
        example with this array:
         [[1, 2], [3, 4]]
        we will generate:
         [ 1, 2, 3, 4 ]
        Individual numbers are represented as doubles.
        Attributes are the way MLIR attaches constant to operations.
        """

        if isinstance(expr, LiteralExprAST):
            return [
                value for inner in expr.values for value in self.collect_data(inner)
            ]
        elif isinstance(expr, NumberExprAST):
            return [expr.val]
        else:
            raise IRGenError(
                f"Unsupported expr ({expr}) of type ({type(expr)}), "
                "expected literal or number expr"
            )

    def ir_gen_call_expr(self, call: CallExprAST) -> SSAValue:
        """
        Emit a call expression. It emits specific operations for the `transpose`
        builtin. Other identifiers are assumed to be user-defined functions.
        """
        assert self.symbol_table is not None
        callee = call.callee

        #    auto location = loc(call.loc());
        # Codegen the operands first.
        operands = [self.ir_gen_expr(expr) for expr in call.args]

        # Builtin calls have their custom operation, meaning this is a
        # straightforward emission.
        if callee == "transpose":
            if len(operands) != 1:
                raise IRGenError(
                    "MLIR codegen encountered an error: toy.transpose "
                    "does not accept multiple arguments"
                )
            op = self.builder.insert(TransposeOp(operands[0]))
            return op.res

        # Otherwise this is a call to a user-defined function. Calls to
        # user-defined functions are mapped to a custom call that takes the callee
        # name as an attribute.
        op = self.builder.insert(
            GenericCallOp(callee, operands, [UnrankedTensorTypeF64(f64)])
        )

        return op.res[0]

    def ir_gen_print_expr(self, call: PrintExprAST):
        """
        Emit a print expression. It emits specific operations for two builtins:
        transpose(x) and print(x).
        """
        arg = self.ir_gen_expr(call.arg)
        self.builder.insert(PrintOp(arg))

    def ir_gen_number_expr(self, num: NumberExprAST) -> SSAValue:
        "Emit a constant for a single number"
        constant_op = self.builder.insert(ConstantOp.from_value(num.val))
        return constant_op.res

    def ir_gen_expr(self, expr: ExprAST) -> SSAValue:
        "Dispatch codegen for the right expression subclass using RTTI."

        if isinstance(expr, BinaryExprAST):
            return self.ir_gen_binary_expr(expr)
        if isinstance(expr, VariableExprAST):
            return self.ir_gen_variable_expr(expr)
        if isinstance(expr, LiteralExprAST):
            return self.ir_gen_literal_expr(expr)
        if isinstance(expr, CallExprAST):
            return self.ir_gen_call_expr(expr)
        if isinstance(expr, NumberExprAST):
            return self.ir_gen_number_expr(expr)
        else:
            raise IRGenError(
                f"MLIR codegen encountered an unhandled expr kind '{type(expr).__name__}'"
            )

    def ir_gen_var_decl_expr(self, vardecl: VarDeclExprAST) -> SSAValue:
        """
        Handle a variable declaration, we'll codegen the expression that forms the
        initializer and record the value in the symbol table before returning it.
        Future expressions will be able to reference this variable through symbol
        table lookup.
        """

        value = self.ir_gen_expr(vardecl.expr)

        # We have the initializer value, but in case the variable was declared
        # with specific shape, we emit a "reshape" operation. It will get
        # optimized out later as needed.
        if len(vardecl.varType.shape):
            reshape_op = self.builder.insert(ReshapeOp(value, vardecl.varType.shape))

            value = reshape_op.res

        # Register the value in the symbol table.
        self.declare(vardecl.name, value)

        return value

    def ir_gen_expr_list(self, exprs: Iterable[ExprAST]) -> None:
        "Codegen a list of expressions, raise error if one of them hit an error."
        assert self.symbol_table is not None

        for expr in exprs:
            # Specific handling for variable declarations, return statement, and
            # print. These can only appear in block list and not in nested
            # expressions.
            if isinstance(expr, VarDeclExprAST):
                self.ir_gen_var_decl_expr(expr)
            elif isinstance(expr, ReturnExprAST):
                self.ir_gen_return_expr(expr)
            elif isinstance(expr, PrintExprAST):
                self.ir_gen_print_expr(expr)
            else:
                # Generic expression dispatch codegen.
                self.ir_gen_expr(expr)



from dataclasses import dataclass

from xdsl.backend.x86.register_allocation import X86RegisterAllocator
from xdsl.backend.x86.register_stack import X86RegisterStack
from xdsl.context import Context
from xdsl.dialects import x86_func
from xdsl.dialects.builtin import ModuleOp
from xdsl.passes import ModulePass


@dataclass(frozen=True)
class X86AllocateRegisters(ModulePass):
    """
    Allocates unallocated registers in the module.
    """

    name = "x86-allocate-registers"

    def apply(self, ctx: Context, op: ModuleOp) -> None:
        for inner_op in op.walk():
            if isinstance(inner_op, x86_func.FuncOp):
                available_registers = X86RegisterStack.get()
                allocator = X86RegisterAllocator(available_registers)
                allocator.allocate_func(inner_op)



from collections.abc import Iterator
from dataclasses import dataclass

from typing_extensions import Self

from xdsl.ir import Block, Operation
from xdsl.traits import IsTerminator


@dataclass(init=False)
class PostOrderIterator(Iterator[Block]):
    """
    Iterates through blocks in a region by a depth first search in post order.
    Each block's successors are processed before the block itself (unless they
    have already been encounted).

    Blocks that are not reachable from the starting block will not appear in the
    iteration.
    """

    stack: list[tuple[Block, bool]]
    seen: set[Block]

    def __init__(self, block: Block) -> None:
        self.stack = [(block, False)]
        self.seen = {block}

    def __iter__(self) -> Self:
        return self

    def __next__(self) -> Block:
        if not self.stack:
            raise StopIteration
        (block, visited) = self.stack.pop()
        while not visited:
            self.stack.append((block, True))
            term = block.last_op
            if isinstance(term, Operation) and term.has_trait(IsTerminator()):
                self.stack.extend(
                    (x, False) for x in reversed(term.successors) if x not in self.seen
                )
                self.seen.update(term.successors)
            # stack cannot be empty here
            (block, visited) = self.stack.pop()
        return block



# RUN: python %s | filecheck %s

from ctypes import c_int32

from xdsl.dialects import arith, builtin
from xdsl.frontend.pyast.context import CodeContext
from xdsl.frontend.pyast.program import FrontendProgram
from xdsl.frontend.pyast.utils.exceptions import CodeGenerationException


def add_i32(operand1: c_int32, operand2: c_int32) -> c_int32: ...


p = FrontendProgram()
p.register_type(c_int32, builtin.i32)
p.register_function(add_i32, arith.AddiOp)
with CodeContext(p):
    # CHECK:      builtin.module {
    # CHECK-NEXT:   func.func @test_add(%x : i32, %y : i32) -> i32 {
    # CHECK-NEXT:     %0 = arith.addi %x, %y : i32
    # CHECK-NEXT:     func.return %0 : i32
    # CHECK-NEXT:   }
    # CHECK-NEXT: }
    def test_add(x: c_int32, y: c_int32) -> c_int32:
        return add_i32(x, operand2=y)


p.compile()
print(p.textual_format())


try:
    with CodeContext(p):
        # CHECK-NEXT: Function arguments must be declared variables.
        def test_args():
            return add_i32(1, 2)  # pyright: ignore[reportArgumentType]

    p.compile(desymref=False)
    print(p.textual_format())
except CodeGenerationException as e:
    print(e.msg)

try:
    with CodeContext(p):
        # CHECK-NEXT: Function arguments must be declared variables.
        def test_args():
            return add_i32(operand1=1, operand2=2)  # pyright: ignore[reportArgumentType]

    p.compile(desymref=False)
    print(p.textual_format())
except CodeGenerationException as e:
    print(e.msg)


def func():
    pass


with CodeContext(p):

    def test_func():
        return func()  # noqa: F821


try:
    # CHECK-NEXT: Function 'func' is not registered.
    p.compile(desymref=False)
    print(p.textual_format())
except CodeGenerationException as e:
    print(e.msg)

try:
    # CHECK-NEXT: Function 'func' is not defined in scope.
    del func
    p.compile(desymref=False)
    print(p.textual_format())
except CodeGenerationException as e:
    print(e.msg)
